# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from typing import Dict, List, Tuple, Union

import gwkokab
from gwkokab.models import NPowerlawMGaussian
from gwkokab.models.utils import create_truncated_normal_distributions
from gwkokab.parameters import Parameters
from kokab.utils.common import expand_arguments
from kokab.utils.f_monk import Monk, monk_arg_parser


class NPowerlawMGaussianMonk(Monk):
    def __init__(
        self,
        N_pl: int,
        N_g: int,
        has_beta_spin: bool,
        has_truncated_normal_spin: bool,
        has_tilt: bool,
        has_eccentricity: bool,
        has_redshift: bool,
        has_cos_iota: bool,
        has_phi_12: bool,
        has_polarization_angle: bool,
        has_right_ascension: bool,
        has_sin_declination: bool,
        has_detection_time: bool,
        data_filename: str,
        seed: int,
        prior_filename: str,
        selection_fn_filename: str,
        poisson_mean_filename: str,
        sampler_settings_filename: str,
        n_samples: int,
        max_iter_mean: int,
        max_iter_cov: int,
        n_vi_steps: int,
        learning_rate: float,
        batch_size: int,
        debug_nans: bool = False,
        profile_memory: bool = False,
        check_leaks: bool = False,
    ) -> None:
        self.N_pl = N_pl
        self.N_g = N_g
        self.has_beta_spin = has_beta_spin
        self.has_truncated_normal_spin = has_truncated_normal_spin
        if self.has_truncated_normal_spin:
            gwkokab.models.npowerlawmgaussian._model.build_spin_distributions = (
                create_truncated_normal_distributions
            )
        self.has_tilt = has_tilt
        self.has_eccentricity = has_eccentricity
        self.has_redshift = has_redshift
        self.has_cos_iota = has_cos_iota
        self.has_phi_12 = has_phi_12
        self.has_polarization_angle = has_polarization_angle
        self.has_right_ascension = has_right_ascension
        self.has_sin_declination = has_sin_declination
        self.has_detection_time = has_detection_time

        super().__init__(
            NPowerlawMGaussian,
            data_filename,
            seed,
            prior_filename,
            selection_fn_filename,
            poisson_mean_filename,
            sampler_settings_filename,
            debug_nans=debug_nans,
            profile_memory=profile_memory,
            check_leaks=check_leaks,
            analysis_name="n_pls_m_gs",
            n_samples=n_samples,
            max_iter_mean=max_iter_mean,
            max_iter_cov=max_iter_cov,
            n_vi_steps=n_vi_steps,
            learning_rate=learning_rate,
            batch_size=batch_size,
        )

    @property
    def constants(self) -> Dict[str, Union[int, float, bool]]:
        return {
            "N_pl": self.N_pl,
            "N_g": self.N_g,
            "use_spin": self.has_beta_spin or self.has_truncated_normal_spin,
            "use_tilt": self.has_tilt,
            "use_eccentricity": self.has_eccentricity,
            "use_redshift": self.has_redshift,
            "use_cos_iota": self.has_cos_iota,
            "use_phi_12": self.has_phi_12,
            "use_polarization_angle": self.has_polarization_angle,
            "use_right_ascension": self.has_right_ascension,
            "use_sin_declination": self.has_sin_declination,
            "use_detection_time": self.has_detection_time,
        }

    @property
    def parameters(self) -> List[str]:
        names = [
            Parameters.PRIMARY_MASS_SOURCE.value,
            Parameters.SECONDARY_MASS_SOURCE.value,
        ]
        if self.has_beta_spin or self.has_truncated_normal_spin:
            names.append(Parameters.PRIMARY_SPIN_MAGNITUDE.value)
            names.append(Parameters.SECONDARY_SPIN_MAGNITUDE.value)
        if self.has_tilt:
            names.extend([Parameters.COS_TILT_1.value, Parameters.COS_TILT_2.value])
        if self.has_phi_12:
            names.append(Parameters.PHI_12.value)
        if self.has_eccentricity:
            names.append(Parameters.ECCENTRICITY.value)
        if self.has_redshift:
            names.append(Parameters.REDSHIFT.value)
        if self.has_right_ascension:
            names.append(Parameters.RIGHT_ASCENSION.value)
        if self.has_sin_declination:
            names.append(Parameters.SIN_DECLINATION.value)
        if self.has_detection_time:
            names.append(Parameters.DETECTION_TIME.value)
        if self.has_cos_iota:
            names.append(Parameters.COS_IOTA.value)
        if self.has_polarization_angle:
            names.append(Parameters.POLARIZATION_ANGLE.value)
        return names

    @property
    def model_parameters(self) -> List[str]:
        all_params: List[Tuple[str, int]] = [
            ("log_rate", self.N_pl + self.N_g),
            ("alpha_pl", self.N_pl),
            ("beta_pl", self.N_pl),
            ("m1_loc_g", self.N_g),
            ("m2_loc_g", self.N_g),
            ("m1_scale_g", self.N_g),
            ("m2_scale_g", self.N_g),
            ("m1_low_g", self.N_g),
            ("m2_low_g", self.N_g),
            ("m1_high_g", self.N_g),
            ("m2_high_g", self.N_g),
            ("mmax_pl", self.N_pl),
            ("mmin_pl", self.N_pl),
        ]

        if self.has_truncated_normal_spin:
            all_params.extend(
                [
                    ("chi1_high_g", self.N_g),
                    ("chi1_high_pl", self.N_pl),
                    ("chi1_loc_g", self.N_g),
                    ("chi1_loc_pl", self.N_pl),
                    ("chi1_low_g", self.N_g),
                    ("chi1_low_pl", self.N_pl),
                    ("chi1_scale_g", self.N_g),
                    ("chi1_scale_pl", self.N_pl),
                    ("chi2_high_g", self.N_g),
                    ("chi2_high_pl", self.N_pl),
                    ("chi2_loc_g", self.N_g),
                    ("chi2_loc_pl", self.N_pl),
                    ("chi2_low_g", self.N_g),
                    ("chi2_low_pl", self.N_pl),
                    ("chi2_scale_g", self.N_g),
                    ("chi2_scale_pl", self.N_pl),
                ]
            )
        if self.has_beta_spin:
            all_params.extend(
                [
                    ("chi1_mean_g", self.N_g),
                    ("chi1_mean_pl", self.N_pl),
                    ("chi1_variance_g", self.N_g),
                    ("chi1_variance_pl", self.N_pl),
                    ("chi2_mean_g", self.N_g),
                    ("chi2_mean_pl", self.N_pl),
                    ("chi2_variance_g", self.N_g),
                    ("chi2_variance_pl", self.N_pl),
                ]
            )

        if self.has_tilt:
            all_params.extend(
                [
                    ("cos_tilt_zeta_g", self.N_g),
                    ("cos_tilt_zeta_pl", self.N_pl),
                    ("cos_tilt1_scale_g", self.N_g),
                    ("cos_tilt1_scale_pl", self.N_pl),
                    ("cos_tilt2_scale_g", self.N_g),
                    ("cos_tilt2_scale_pl", self.N_pl),
                ]
            )

        if self.has_phi_12:
            all_params.extend(
                [
                    (Parameters.PHI_12.value + "_high_g", self.N_g),
                    (Parameters.PHI_12.value + "_high_pl", self.N_pl),
                    (Parameters.PHI_12.value + "_loc_g", self.N_g),
                    (Parameters.PHI_12.value + "_loc_pl", self.N_pl),
                    (Parameters.PHI_12.value + "_low_g", self.N_g),
                    (Parameters.PHI_12.value + "_low_pl", self.N_pl),
                    (Parameters.PHI_12.value + "_scale_g", self.N_g),
                    (Parameters.PHI_12.value + "_scale_pl", self.N_pl),
                ]
            )

        if self.has_eccentricity:
            all_params.extend(
                [
                    ("ecc_high_g", self.N_g),
                    ("ecc_high_pl", self.N_pl),
                    ("ecc_loc_g", self.N_g),
                    ("ecc_loc_pl", self.N_pl),
                    ("ecc_low_g", self.N_g),
                    ("ecc_low_pl", self.N_pl),
                    ("ecc_scale_g", self.N_g),
                    ("ecc_scale_pl", self.N_pl),
                ]
            )

        if self.has_redshift:
            all_params.extend(
                [
                    (Parameters.REDSHIFT.value + "_kappa_g", self.N_g),
                    (Parameters.REDSHIFT.value + "_kappa_pl", self.N_pl),
                    (Parameters.REDSHIFT.value + "_z_max_g", self.N_g),
                    (Parameters.REDSHIFT.value + "_z_max_pl", self.N_pl),
                ]
            )

        if self.has_right_ascension:
            all_params.extend(
                [
                    (Parameters.RIGHT_ASCENSION.value + "_high_g", self.N_g),
                    (Parameters.RIGHT_ASCENSION.value + "_high_pl", self.N_pl),
                    (Parameters.RIGHT_ASCENSION.value + "_loc_g", self.N_g),
                    (Parameters.RIGHT_ASCENSION.value + "_loc_pl", self.N_pl),
                    (Parameters.RIGHT_ASCENSION.value + "_low_g", self.N_g),
                    (Parameters.RIGHT_ASCENSION.value + "_low_pl", self.N_pl),
                    (Parameters.RIGHT_ASCENSION.value + "_scale_g", self.N_g),
                    (Parameters.RIGHT_ASCENSION.value + "_scale_pl", self.N_pl),
                ]
            )

        if self.has_sin_declination:
            all_params.extend(
                [
                    (Parameters.SIN_DECLINATION.value + "_high_g", self.N_g),
                    (Parameters.SIN_DECLINATION.value + "_high_pl", self.N_pl),
                    (Parameters.SIN_DECLINATION.value + "_loc_g", self.N_g),
                    (Parameters.SIN_DECLINATION.value + "_loc_pl", self.N_pl),
                    (Parameters.SIN_DECLINATION.value + "_low_g", self.N_g),
                    (Parameters.SIN_DECLINATION.value + "_low_pl", self.N_pl),
                    (Parameters.SIN_DECLINATION.value + "_scale_g", self.N_g),
                    (Parameters.SIN_DECLINATION.value + "_scale_pl", self.N_pl),
                ]
            )

        if self.has_detection_time:
            all_params.extend(
                [
                    (Parameters.DETECTION_TIME.value + "_high_g", self.N_g),
                    (Parameters.DETECTION_TIME.value + "_high_pl", self.N_pl),
                    (Parameters.DETECTION_TIME.value + "_low_g", self.N_g),
                    (Parameters.DETECTION_TIME.value + "_low_pl", self.N_pl),
                ]
            )

        if self.has_cos_iota:
            all_params.extend(
                [
                    (Parameters.COS_IOTA.value + "_high_g", self.N_g),
                    (Parameters.COS_IOTA.value + "_high_pl", self.N_pl),
                    (Parameters.COS_IOTA.value + "_loc_g", self.N_g),
                    (Parameters.COS_IOTA.value + "_loc_pl", self.N_pl),
                    (Parameters.COS_IOTA.value + "_low_g", self.N_g),
                    (Parameters.COS_IOTA.value + "_low_pl", self.N_pl),
                    (Parameters.COS_IOTA.value + "_scale_g", self.N_g),
                    (Parameters.COS_IOTA.value + "_scale_pl", self.N_pl),
                ]
            )

        if self.has_polarization_angle:
            all_params.extend(
                [
                    (Parameters.POLARIZATION_ANGLE.value + "_high_g", self.N_g),
                    (Parameters.POLARIZATION_ANGLE.value + "_high_pl", self.N_pl),
                    (Parameters.POLARIZATION_ANGLE.value + "_loc_g", self.N_g),
                    (Parameters.POLARIZATION_ANGLE.value + "_loc_pl", self.N_pl),
                    (Parameters.POLARIZATION_ANGLE.value + "_low_g", self.N_g),
                    (Parameters.POLARIZATION_ANGLE.value + "_low_pl", self.N_pl),
                    (Parameters.POLARIZATION_ANGLE.value + "_scale_g", self.N_g),
                    (Parameters.POLARIZATION_ANGLE.value + "_scale_pl", self.N_pl),
                ]
            )

        extended_params = []
        for params in all_params:
            extended_params.extend(expand_arguments(*params))
        return extended_params


def main() -> None:
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser = monk_arg_parser(parser)

    model_group = parser.add_argument_group("Model Options")
    model_group.add_argument(
        "--n-pl",
        type=int,
        help="Number of power-law components in the mass model.",
    )
    model_group.add_argument(
        "--n-g",
        type=int,
        help="Number of Gaussian components in the mass model.",
    )

    spin_group = model_group.add_mutually_exclusive_group()
    spin_group.add_argument(
        "--add-beta-spin",
        action="store_true",
        help="Include beta spin parameters in the model.",
    )
    spin_group.add_argument(
        "--add-truncated-normal-spin",
        action="store_true",
        help="Include truncated normal spin parameters in the model.",
    )

    model_group.add_argument(
        "--add-tilt",
        action="store_true",
        help="Include tilt parameters in the model.",
    )
    model_group.add_argument(
        "--add-redshift",
        action="store_true",
        help="Include redshift parameter in the model",
    )
    model_group.add_argument(
        "--add-truncated-normal-eccentricity",
        action="store_true",
        help="Include truncated normal eccentricity in the model.",
    )
    model_group.add_argument(
        "--add-cos-iota",
        action="store_true",
        help="Include cos_iota parameter in the model",
    )
    model_group.add_argument(
        "--add-phi-12",
        action="store_true",
        help="Include phi_12 parameter in the model",
    )
    model_group.add_argument(
        "--add-polarization-angle",
        action="store_true",
        help="Include polarization_angle parameter in the model",
    )
    model_group.add_argument(
        "--add-right-ascension",
        action="store_true",
        help="Include right_ascension parameter in the model",
    )
    model_group.add_argument(
        "--add-sin-declination",
        action="store_true",
        help="Include sin_declination parameter in the model",
    )
    model_group.add_argument(
        "--add-detection-time",
        action="store_true",
        help="Include detection_time parameter in the model",
    )

    args = parser.parse_args()

    NPowerlawMGaussianMonk(
        N_pl=args.n_pl,
        N_g=args.n_g,
        has_beta_spin=args.add_beta_spin,
        has_truncated_normal_spin=args.add_truncated_normal_spin,
        has_tilt=args.add_tilt,
        has_eccentricity=args.add_truncated_normal_eccentricity,
        has_redshift=args.add_redshift,
        has_cos_iota=args.add_cos_iota,
        has_phi_12=args.add_phi_12,
        has_polarization_angle=args.add_polarization_angle,
        has_right_ascension=args.add_right_ascension,
        has_sin_declination=args.add_sin_declination,
        has_detection_time=args.add_detection_time,
        data_filename=args.data_filename,
        seed=args.seed,
        prior_filename=args.prior_json,
        selection_fn_filename=args.vt_json,
        poisson_mean_filename=args.pmean_json,
        sampler_settings_filename=args.sampler_config,
        n_samples=args.n_samples,
        max_iter_mean=args.max_iter_mean,
        max_iter_cov=args.max_iter_cov,
        n_vi_steps=args.n_vi_steps,
        learning_rate=args.learning_rate,
        batch_size=args.batch_size,
        debug_nans=args.debug_nans,
        profile_memory=args.profile_memory,
        check_leaks=args.check_leaks,
    ).run()



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


# Copyright (c) 2023 Farr Out Lab
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from ._cosmology import Cosmology


PLANCK_2015_Ho = 67.74 * 1e3  # m/s/Mpc
PLANCK_2015_OmegaMatter = 0.3089
PLANCK_2015_OmegaLambda = 1.0 - PLANCK_2015_OmegaMatter
PLANCK_2015_OmegaRadiation = 0.0

PLANCK_2015_Cosmology = Cosmology(
    PLANCK_2015_Ho,
    PLANCK_2015_OmegaMatter,
    PLANCK_2015_OmegaRadiation,
    PLANCK_2015_OmegaLambda,
)
"Cosmology : See Table4 in arXiv:1502.01589, OmegaMatter from astropy Planck 2015"

PLANCK_2018_Ho = 67.32 * 1e3
PLANCK_2018_OmegaMatter = 0.3158
PLANCK_2018_OmegaLambda = 1.0 - PLANCK_2018_OmegaMatter
PLANCK_2018_OmegaRadiation = 0.0

PLANCK_2018_Cosmology = Cosmology(
    PLANCK_2018_Ho,
    PLANCK_2018_OmegaMatter,
    PLANCK_2018_OmegaRadiation,
    PLANCK_2018_OmegaLambda,
)
"Cosmology : See Table1 in arXiv:1807.06209"



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from collections.abc import Callable, Sequence
from typing import Optional

import equinox as eqx
import jax
from jax import lax, numpy as jnp
from jaxtyping import Array
from loguru import logger

from ..utils.tools import error_if
from ._abc import VolumeTimeSensitivityInterface
from ._utils import load_model


class NeuralNetVolumeTimeSensitivity(VolumeTimeSensitivityInterface):
    neural_vt_model: eqx.nn.MLP
    """The neural volume-time sensitivity model."""

    def __init__(
        self,
        parameters: Sequence[str],
        filename: str,
        batch_size: Optional[int] = None,
    ) -> None:
        """Convenience class for loading a neural vt.

        Parameters
        ----------
        parameters : Sequence[str]
            The names of the parameters that the model expects.
        filename : str
            The filename of the neural vt.
        batch_size : Optional[int], optional
            The batch size :func:`jax.lax.map` should use, by default None.
        """
        logger.debug(
            "Starting `NeuralNetVolumeTimeSensitivity` from {} with parameters {}",
            filename,
            parameters,
        )
        error_if(not parameters, msg="parameters sequence cannot be empty")
        error_if(
            not isinstance(parameters, Sequence),
            err=TypeError,
            msg=f"parameters must be a Sequence, got {type(parameters)}",
        )
        error_if(
            not all(isinstance(p, str) for p in parameters),
            err=TypeError,
            msg="all parameters must be strings",
        )
        if batch_size is not None:
            error_if(
                not isinstance(batch_size, int),
                err=TypeError,
                msg=f"batch_size must be an integer, got {type(batch_size)}",
            )
            error_if(
                batch_size < 1,
                msg=f"batch_size must be a positive integer, got {batch_size}",
            )

        self.batch_size = batch_size
        logger.debug("Batch size set to {}", self.batch_size)
        names, self.neural_vt_model = load_model(filename)
        error_if(
            any(name not in parameters for name in names),
            msg=f"Model in {filename} expects parameters {names}, but received "
            f"{parameters}. Missing: {set(names) - set(parameters)}",
        )
        self.shuffle_indices = [parameters.index(name) for name in names]
        logger.debug("Finished `NeuralNetVolumeTimeSensitivity`")

    def get_logVT(self) -> Callable[[Array], Array]:
        """Gets the logVT function."""

        @jax.jit
        def _logVT(x: Array) -> Array:
            x_new = x[..., self.shuffle_indices]
            return self.neural_vt_model(x_new)

        return _logVT

    def get_mapped_logVT(self) -> Callable[[Array], Array]:
        """Gets the vmapped logVT function for batch processing."""
        _batch_size = self.batch_size

        @jax.jit
        def _logVT(x: Array) -> Array:
            x_new = x[..., self.shuffle_indices]
            return jnp.squeeze(
                lax.map(self.neural_vt_model, x_new, batch_size=_batch_size), axis=-1
            )

        return _logVT



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from typing import Callable, Dict, List, Literal, Optional, Tuple

from jax import numpy as jnp, tree as jtr
from jaxtyping import Array
from numpyro.distributions import Distribution

from ..constraints import any_constraint
from ..utils import (
    ExtendedSupportTransformedDistribution,
    JointDistribution,
    ScaledMixture,
)
from ._ncombination import (
    combine_distributions,
    create_beta_distributions,
    create_independent_spin_orientation_gaussian_isotropic,
    create_powerlaw_redshift,
    create_powerlaws,
    create_truncated_normal_distributions,
    create_uniform_distributions,
)


build_spin_distributions = create_beta_distributions
build_tilt_distributions = create_independent_spin_orientation_gaussian_isotropic
build_eccentricity_distributions = create_truncated_normal_distributions
build_redshift_distributions = create_powerlaw_redshift
build_cos_iota_distribution = create_uniform_distributions
build_phi_12_distribution = create_uniform_distributions
build_polarization_angle_distribution = create_uniform_distributions
build_right_ascension_distribution = create_uniform_distributions
build_sin_declination_distribution = create_uniform_distributions
build_mean_anomaly_distribution = create_uniform_distributions
build_detection_time_distribution = create_uniform_distributions
build_phi_1_distribution = create_uniform_distributions
build_phi_2_distribution = create_uniform_distributions
build_phi_orb_distribution = create_uniform_distributions


def _build_non_mass_distributions(
    N: int,
    component_type: Literal["pl", "g"],
    mass_distributions: List[Distribution],
    use_spin: bool,
    use_tilt: bool,
    use_eccentricity: bool,
    use_mean_anomaly: bool,
    use_redshift: bool,
    use_cos_iota: bool,
    use_polarization_angle: bool,
    use_right_ascension: bool,
    use_sin_declination: bool,
    use_detection_time: bool,
    use_phi_1: bool,
    use_phi_2: bool,
    use_phi_12: bool,
    use_phi_orb: bool,
    params: Dict[str, Array],
    validate_args: Optional[bool] = None,
) -> List[Distribution]:
    """Build distributions for non-mass parameters.

    Parameters
    ----------
    N : int
        Number of components
    component_type : Literal[&quot;pl&quot;, &quot;g&quot;]
        type of component, either "pl" or "g"
    mass_distributions : List[Distribution]
        list of mass distributions
    use_spin : bool
        whether to include spin
    use_tilt : bool
        whether to include tilt
    use_eccentricity : bool
        whether to include eccentricity
    use_mean_anomaly : bool
        whether to include mean_anomaly
    use_redshift : bool
        whether to include redshift
    use_cos_iota : bool
        whether to include cos_iota
    use_polarization_angle : bool
        whether to include polarization_angle
    use_right_ascension : bool
        whether to include right_ascension
    use_sin_declination : bool
        whether to include sin_declination
    use_detection_time : bool
        whether to include detection_time
    use_phi_1 : bool
        whether to include phi_1
    use_phi_2 : bool
        whether to include phi_2
    use_phi_12 : bool
        whether to include phi_12
    use_phi_orb : bool
        whether to include phi_orb
    params : Dict[str, Array]
        dictionary of parameters
    validate_args : Optional[bool], optional
        whether to validate arguments, by default None

    Returns
    -------
    List[Distribution]
        list of distributions
    """
    build_distributions = mass_distributions
    _info_collection: List[Tuple[bool, str, Callable[..., List[Distribution]]]] = [
        (use_spin, "chi1", build_spin_distributions),
        (use_spin, "chi2", build_spin_distributions),
        # combined tilt distribution
        (use_tilt, "cos_tilt1_cos_tilt2", build_tilt_distributions),
        (use_phi_1, "phi_1", build_phi_1_distribution),
        (use_phi_2, "phi_2", build_phi_2_distribution),
        (use_phi_12, "phi_12", build_phi_12_distribution),
        (use_eccentricity, "ecc", build_eccentricity_distributions),
        (use_mean_anomaly, "mean_anomaly", build_mean_anomaly_distribution),
        (use_redshift, "redshift", build_redshift_distributions),
        (use_right_ascension, "ra", build_right_ascension_distribution),
        (use_sin_declination, "dec", build_sin_declination_distribution),
        (use_detection_time, "detection_time", build_detection_time_distribution),
        (use_cos_iota, "cos_iota", build_cos_iota_distribution),
        (use_polarization_angle, "psi", build_polarization_angle_distribution),
        (use_phi_orb, "phi_orb", build_phi_orb_distribution),
    ]

    # Iterate over the list of tuples and build distributions
    for use, param_name, build_func in _info_collection:
        if use:
            distributions = build_func(
                N=N,
                parameter_name=param_name,
                component_type=component_type,
                params=params,
                validate_args=validate_args,
            )
            build_distributions = combine_distributions(
                build_distributions, distributions
            )

    return build_distributions


def _build_pl_component_distributions(
    N: int,
    use_spin: bool,
    use_tilt: bool,
    use_eccentricity: bool,
    use_mean_anomaly: bool,
    use_redshift: bool,
    use_cos_iota: bool,
    use_polarization_angle: bool,
    use_right_ascension: bool,
    use_sin_declination: bool,
    use_detection_time: bool,
    use_phi_1: bool,
    use_phi_2: bool,
    use_phi_12: bool,
    use_phi_orb: bool,
    params: Dict[str, Array],
    validate_args: Optional[bool] = None,
) -> List[JointDistribution]:
    """Build distributions for power-law components.

    Parameters
    ----------
    N : int
        Number of components
    use_spin : bool
        whether to include spin
    use_tilt : bool
        whether to include tilt
    use_eccentricity : bool
        whether to include eccentricity
    use_mean_anomaly : bool
        whether to include mean_anomaly
    use_redshift : bool
        whether to include redshift
    use_cos_iota : bool
        whether to include cos_iota
    use_polarization_angle : bool
        whether to include polarization_angle
    use_right_ascension : bool
        whether to include right_ascension
    use_sin_declination : bool
        whether to include sin_declination
    use_detection_time : bool
        whether to include detection_time
    use_phi_1 : bool
        whether to include phi_1
    use_phi_2 : bool
        whether to include phi_2
    use_phi_12 : bool
        whether to include phi_12
    use_phi_orb : bool
        whether to include phi_orb
    params : Dict[str, Array]
        dictionary of parameters
    validate_args : Optional[bool], optional
        whether to validate arguments, by default None

    Returns
    -------
    List[JointDistribution]
        list of JointDistribution
    """
    powerlaws = create_powerlaws(N=N, params=params, validate_args=validate_args)

    mass_distributions = jtr.map(
        lambda powerlaw: [powerlaw],
        powerlaws,
        is_leaf=lambda x: isinstance(x, ExtendedSupportTransformedDistribution),
    )

    build_distributions = _build_non_mass_distributions(
        N=N,
        component_type="pl",
        mass_distributions=mass_distributions,
        use_spin=use_spin,
        use_tilt=use_tilt,
        use_eccentricity=use_eccentricity,
        use_redshift=use_redshift,
        use_cos_iota=use_cos_iota,
        use_phi_12=use_phi_12,
        use_polarization_angle=use_polarization_angle,
        use_right_ascension=use_right_ascension,
        use_sin_declination=use_sin_declination,
        use_detection_time=use_detection_time,
        use_phi_1=use_phi_1,
        use_phi_2=use_phi_2,
        use_phi_orb=use_phi_orb,
        use_mean_anomaly=use_mean_anomaly,
        params=params,
        validate_args=validate_args,
    )

    return [
        JointDistribution(*dists, validate_args=validate_args)
        for dists in build_distributions
    ]


def _build_g_component_distributions(
    N: int,
    use_spin: bool,
    use_tilt: bool,
    use_eccentricity: bool,
    use_mean_anomaly: bool,
    use_redshift: bool,
    use_cos_iota: bool,
    use_polarization_angle: bool,
    use_right_ascension: bool,
    use_sin_declination: bool,
    use_detection_time: bool,
    use_phi_1: bool,
    use_phi_2: bool,
    use_phi_12: bool,
    use_phi_orb: bool,
    params: Dict[str, Array],
    validate_args: Optional[bool] = None,
) -> List[JointDistribution]:
    """Build distributions for Gaussian components.

    Parameters
    ----------
    N : int
        Number of components
    use_spin : bool
        whether to include spin
    use_tilt : bool
        whether to include tilt
    use_eccentricity : bool
        whether to include eccentricity
    use_mean_anomaly : bool
        whether to include mean_anomaly
    use_redshift : bool
        whether to include redshift
    use_cos_iota : bool
        whether to include cos_iota
    use_polarization_angle : bool
        whether to include polarization_angle
    use_right_ascension : bool
        whether to include right_ascension
    use_sin_declination : bool
        whether to include sin_declination
    use_detection_time : bool
        whether to include detection_time
    use_phi_1 : bool
        whether to include phi_1
    use_phi_2 : bool
        whether to include phi_2
    use_phi_12 : bool
        whether to include phi_12
    use_phi_orb : bool
        whether to include phi_orb
    params : Dict[str, Array]
        dictionary of parameters
    validate_args : Optional[bool], optional
        whether to validate arguments, by default None

    Returns
    -------
    List[JointDistribution]
        list of JointDistribution
    """
    m1_dists = create_truncated_normal_distributions(
        N=N,
        parameter_name="m1",
        component_type="g",
        params=params,
        validate_args=validate_args,
    )
    m2_dists = create_truncated_normal_distributions(
        N=N,
        parameter_name="m2",
        component_type="g",
        params=params,
        validate_args=validate_args,
    )

    mass_distributions = jtr.map(
        lambda m1, m2: [m1, m2],
        m1_dists,
        m2_dists,
        is_leaf=lambda x: isinstance(x, Distribution),
    )

    build_distributions = _build_non_mass_distributions(
        N=N,
        component_type="g",
        mass_distributions=mass_distributions,
        use_spin=use_spin,
        use_tilt=use_tilt,
        use_eccentricity=use_eccentricity,
        use_redshift=use_redshift,
        use_cos_iota=use_cos_iota,
        use_phi_12=use_phi_12,
        use_polarization_angle=use_polarization_angle,
        use_right_ascension=use_right_ascension,
        use_sin_declination=use_sin_declination,
        use_detection_time=use_detection_time,
        use_phi_1=use_phi_1,
        use_phi_2=use_phi_2,
        use_phi_orb=use_phi_orb,
        use_mean_anomaly=use_mean_anomaly,
        params=params,
        validate_args=validate_args,
    )

    return [
        JointDistribution(*dists, validate_args=validate_args)
        for dists in build_distributions
    ]


def NPowerlawMGaussian(
    N_pl: int,
    N_g: int,
    use_spin: bool = False,
    use_tilt: bool = False,
    use_eccentricity: bool = False,
    use_redshift: bool = False,
    use_cos_iota: bool = False,
    use_phi_12: bool = False,
    use_polarization_angle: bool = False,
    use_right_ascension: bool = False,
    use_sin_declination: bool = False,
    use_detection_time: bool = False,
    use_phi_1: bool = False,
    use_phi_2: bool = False,
    use_phi_orb: bool = False,
    use_mean_anomaly: bool = False,
    *,
    validate_args=None,
    **params,
) -> ScaledMixture:
    r"""Create a mixture of power-law and Gaussian components.

    This model has a lot of parameters, we can not list all of them here. Therefore,
    we are providing the general description of the each sub model and their parameters.

    .. important::

        Important information about this models are as follows:

        * The first :code:`N_pl` components are power-law and the next :code:`N_g` components are Gaussian.
        * Log rates are named as :code:`log_rate_{i}` where :code:`i` is the index of the component.
        * First :code:`N_pl` log rates are for power-law components and the next :code:`N_g` log rates are for Gaussian components.
        * All log rates are in terms of natural logarithm.


    .. note::

        **Mass distribution**: For powerlaw mass distribution is
        :class:`PowerlawPrimaryMassRatio` and for Gaussian we have
        :class:`~numpyro.distributions.truncated.TruncatedNormal` distribution.

        .. math:: (m_1, m_2) \sim \text{PowerlawPrimaryMassRatio}(\alpha, \beta, m_{\text{min}}, m_{\text{max}})

        .. math:: (m_1, m_2) \sim \mathcal{N}_{[a,b]}(\mu, \sigma^2)

        **Spin distribution**: Spin is taken from a beta distribution. The beta distribution
        is parameterized by :math:`\mu` and :math:`\sigma^2` where :math:`\mu` is the mean
        and :math:`\sigma^2` is the variance of the beta distribution.

        .. math:: \chi_i \sim \mathrm{Beta}(\mu, \sigma^2)

        .. warning::
            Not every choice of :math:`\mu` and :math:`\sigma^2` will result in a valid beta
            distribution. The beta distribution is only valid when :math:`\mu \in (0, 1)`,
            :math:`\sigma^2 \in (0, 0.25)`, and :math:`\mu(1-\mu) > \sigma^2`. Refer to the
            `link <https://stats.stackexchange.com/questions/12232/calculating-the-parameters-of-a-beta-distribution-using-the-mean-and-variance#comment926602_12239>`_
            for more information.

        **Tilt distribution**: Tilt is taken from a
        :class:`~numpyro.distributions.truncated.TruncatedNormal` distribution, with
        fixed mean :math:`\mu=1` and fixed bounds, :math:`a=-1` and :math:`b=1`.

        .. math:: \cos(\theta_i) \sim \mathcal{N}_{[-1, 1]}(\sigma^2\mid\mu=1)

        **Eccentricity distribution**: Eccentricity is taken from
        :class:`~numpyro.distributions.truncated.TruncatedNormal` distribution.

        .. math:: \varepsilon_i \sim \mathcal{N}_{[a,b]}(\mu, \sigma^2)


    .. attention::

        Interestingly, in :class:`~numpyro.distributions.truncated.TruncatedNormal`
        distribution, if any of the bounds are not provided, it will be set to
        :math:`\pm\infty`. For example, if we set :math:`\mu=0` and do not provide the
        upper bound, then the resulting distribution would be a
        `half normal <https://en.wikipedia.org/wiki/Half-normal_distribution>`_
        distribution.


    The naming of the parameters follows the following convention:

    .. code:: bash

        <parameter name>_<model parameter>_<component type>_<component number>

    with an exception for the powerlaw mass distribution where the
    :code:`<parameter name>` is ignored. For example, spin is taken from a beta
    distribution whose parameters are :code:`mean` and :code:`variance`. The naming
    convention for the spin parameters would be:

    .. code:: text

        chi[1-2]_mean_(pl|g)_[0-N_pl+N_g]
        chi[1-2]_variance_(pl|g)_[0-N_pl+N_g]

    Parameters
    ----------

    N_pl : int
        Number of power-law components
    N_g : int
        Number of Gaussian components
    use_spin : bool
        whether to include spin, defaults to False
    use_tilt : bool
        whether to include tilt, defaults to False
    use_eccentricity : bool
        whether to include eccentricity, defaults to False
    use_mean_anomaly : bool
        whether to include mean_anomaly, defaults to False
    use_redshift : bool
        whether to include redshift, defaults to False
    use_cos_iota : bool
        whether to include cos_iota, defaults to False
    use_polarization_angle : bool
        whether to include polarization_angle, defaults to False
    use_right_ascension : bool
        whether to include right_ascension, defaults to False
    use_sin_declination : bool
        whether to include sin_declination, defaults to False
    use_detection_time : bool
        whether to include detection_time, defaults to False
    use_phi_1 : bool
        whether to include phi_1, defaults to False
    use_phi_2 : bool
        whether to include phi_2, defaults to False
    use_phi_12 : bool
        whether to include phi_12, defaults to False
    use_phi_orb : bool
        whether to include phi_orb, defaults to False
    validate_args : Optional[bool], optional
        whether to validate arguments, defaults to None

    Returns
    -------
    ScaledMixture
        scaled mixture of distributions
    """
    if N_pl > 0:
        pl_component_dist = _build_pl_component_distributions(
            N=N_pl,
            use_spin=use_spin,
            use_tilt=use_tilt,
            use_eccentricity=use_eccentricity,
            use_redshift=use_redshift,
            use_cos_iota=use_cos_iota,
            use_phi_12=use_phi_12,
            use_polarization_angle=use_polarization_angle,
            use_right_ascension=use_right_ascension,
            use_sin_declination=use_sin_declination,
            use_detection_time=use_detection_time,
            use_phi_1=use_phi_1,
            use_phi_2=use_phi_2,
            use_phi_orb=use_phi_orb,
            use_mean_anomaly=use_mean_anomaly,
            params=params,
            validate_args=validate_args,
        )

    if N_g > 0:
        g_component_dist = _build_g_component_distributions(
            N=N_g,
            use_spin=use_spin,
            use_tilt=use_tilt,
            use_eccentricity=use_eccentricity,
            use_redshift=use_redshift,
            use_cos_iota=use_cos_iota,
            use_phi_12=use_phi_12,
            use_polarization_angle=use_polarization_angle,
            use_right_ascension=use_right_ascension,
            use_sin_declination=use_sin_declination,
            use_detection_time=use_detection_time,
            use_phi_1=use_phi_1,
            use_phi_2=use_phi_2,
            use_phi_orb=use_phi_orb,
            use_mean_anomaly=use_mean_anomaly,
            params=params,
            validate_args=validate_args,
        )

    if N_pl == 0 and N_g != 0:
        component_dists = g_component_dist
    elif N_g == 0 and N_pl != 0:
        component_dists = pl_component_dist
    else:
        component_dists = pl_component_dist + g_component_dist

    N = N_pl + N_g
    log_rates = jnp.stack([params[f"log_rate_{i}"] for i in range(N)], axis=-1)

    return ScaledMixture(
        log_rates,
        component_dists,
        support=any_constraint(
            [component_dists.support for component_dists in component_dists]
        ),
        validate_args=validate_args,
    )



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import argparse
from typing import Any, List, Tuple

import h5py
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from kokab.utils.ppd import get_all_marginals


def make_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="PPD plot",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        epilog="This script plots ppd plots.",
    )
    parser.add_argument(
        "--data",
        nargs="+",
        help="data files paths (one or more). Only .hdf5 files are supported.",
        type=str,
        required=True,
    )
    parser.add_argument(
        "--dir",
        type=str,
        required=True,
        help="directory to save the plots",
    )
    parser.add_argument(
        "--prefix",
        help="prefix for the output file",
        type=str,
    )

    pretty_group = parser.add_argument_group("Pretty Options")
    pretty_group.add_argument(
        "--titles",
        nargs="+",
        type=str,
        help="titles for the plots in order of the parameters",
    )
    pretty_group.add_argument(
        "--x-scale",
        help="scale of the x-axis",
        default="linear",
        choices=["linear", "log"],
        type=str,
    )
    pretty_group.add_argument(
        "--y-scale",
        help="scale of the y-axis",
        default="linear",
        choices=["linear", "log"],
        type=str,
    )
    pretty_group.add_argument(
        "--x-range",
        nargs=2,
        action="append",
        type=float,
        help="range of the x axis plot in the form of start end for each parameter",
    )
    pretty_group.add_argument(
        "--y-range",
        nargs=2,
        action="append",
        type=float,
        help="range of the y axis plot in the form of start end for each parameter",
    )
    pretty_group.add_argument(
        "--font-size",
        default=16,
        type=int,
        help="font size for the x,y axis labels and titles",
    )
    pretty_group.add_argument(
        "--x-labels",
        nargs="+",
        type=str,
        help="labels for the x axis in order of the parameters",
    )
    pretty_group.add_argument(
        "--y-labels",
        nargs="+",
        type=str,
        help="labels for the y axis in order of the parameters",
    )
    pretty_group.add_argument(
        "--size",
        help="size of the ppd plot in inches",
        nargs=2,
        default=(12, 12),
        type=float,
    )
    pretty_group.add_argument(
        "--use-latex",
        help="use LaTeX for rendering text",
        action="store_true",
    )
    pretty_group.add_argument(
        "--font-family",
        help="font family to use",
        type=str,
        default=None,
    )
    pretty_group.add_argument(
        "--median-color",
        help="color of the median line",
        type=str,
        default=r"#7B8794",
    )
    pretty_group.add_argument(
        "--ppd-color",
        help="color of the ppd line",
        type=str,
        default=r"#b41f78",
    )
    pretty_group.add_argument(
        "--ninety-ci-color",
        help="color of the 90 percentile CI",
        type=str,
        default=r"#C5C9C7",
    )
    pretty_group.add_argument(
        "--fifty-ci-color",
        help="color of the 50 percentile CI",
        type=str,
        default=r"#BBF90F",
    )
    pretty_group.add_argument(
        "--median-alpha",
        help="alpha of the median line",
        type=float,
        default=0.8,
    )
    pretty_group.add_argument(
        "--ppd-alpha",
        help="alpha of the ppd line",
        type=float,
        default=1.0,
    )
    pretty_group.add_argument(
        "--ninety-ci-alpha",
        help="alpha of the 90 percentile CI",
        type=float,
        default=0.5,
    )
    pretty_group.add_argument(
        "--fifty-ci-alpha",
        help="alpha of the 50 percentile CI",
        type=float,
        default=0.7,
    )
    pretty_group.add_argument(
        "--median-linestyle",
        help="linestyle of the median line",
        type=str,
        default="--",
    )
    pretty_group.add_argument(
        "--grid",
        help="show grid",
        action="store_true",
    )
    pretty_group.add_argument(
        "--grid-which",
        help="which grid to show",
        type=str,
        default="both",
        choices=["both", "major", "minor"],
    )
    pretty_group.add_argument(
        "--grid-alpha",
        help="alpha of the grid",
        type=float,
        default=0.7,
    )
    pretty_group.add_argument(
        "--grid-linestyle",
        help="linestyle of the grid",
        type=str,
        default="--",
    )
    parser.add_argument(
        "--dpi",
        help="dots per inch to save file",
        type=int,
        default=100,
    )

    return parser


def get_domain(
    domain_as_read: Any,
) -> List[Tuple[float, float, int]]:
    """Get the domain of the model per axis.

    Parameters
    ----------
    domain_as_read : Any
        The domain of the model per axis.

    Returns
    -------
    List[Tuple[float, float, int]]
        The domain of the model per axis.
    """
    return [(float(s), float(e), int(n)) for s, e, n in domain_as_read]


def get_utf8_decoded_headers(headers_as_read: Any) -> List[str]:
    """Get utf-8 decoded headers.

    Parameters
    ----------
    headers_as_read : Any
        The headers as read from the file.

    Returns
    -------
    List[str]
        The utf-8 decoded headers.
    """
    return [h.decode("utf-8") for h in headers_as_read]


def get_quantiles(data: np.ndarray, quantiles: np.ndarray) -> np.ndarray:
    """Get the quantiles of the data.

    Parameters
    ----------
    data : np.ndarray
        The data to get the quantiles of.
    quantiles : np.ndarray
        The quantiles to get. The quantiles should be between 0 and 1.

    Returns
    -------
    np.ndarray
        The quantiles of the data.
    """
    return np.quantile(data, quantiles, axis=1)


def main() -> None:
    """Main function of the script."""
    parser = make_parser()
    args = parser.parse_args()

    plt.rcParams.update({"text.usetex": args.use_latex})
    if args.font_family is not None:
        plt.rcParams.update({"font.family": args.font_family})

    prefix = "" if args.prefix is None else args.prefix

    x_range = args.x_range
    y_range = args.y_range
    font_size = args.font_size
    x_labels = args.x_labels
    y_labels = args.y_labels
    titles = args.titles

    quantiles = [
        0.05,  # 90% CI
        0.25,  # 50% CI
        0.5,  # median
        0.75,  # 50% CI
        0.95,  # 90% CI
    ]

    axes: List[Tuple[Figure, Axes]] = []

    for j in range(len(args.data)):
        with h5py.File(args.data[j], "r") as f:
            domains = get_domain(f["domains"])
            headers = get_utf8_decoded_headers(f["headers"])
            marginals = [f["marginals"][head][:] for head in headers]
            ppd = f["ppd"][()]

        marginal_ppds = get_all_marginals(ppd, domains)

        if not axes:
            axes = [plt.subplots(figsize=args.size) for _ in range(len(headers))]

        for i in range(len(headers)):
            _, ax = axes[i]
            head = headers[i]
            marginal = marginals[i]
            marginal_ppd = marginal_ppds[i]
            domain = domains[i]

            start, end, num_points = domain
            quant = get_quantiles(marginal, quantiles)
            xx = np.linspace(start, end, num_points)

            ax.plot(
                xx,
                quant[2],
                # label="Median",
                color=args.median_color,
                alpha=args.median_alpha,
                linestyle=args.median_linestyle,
            )
            ax.plot(
                xx,
                marginal_ppd,
                # label="PPD",
                color=args.ppd_color,
                alpha=args.ppd_alpha,
            )
            ax.fill_between(
                xx,
                quant[0],
                quant[-1],
                alpha=args.ninety_ci_alpha,
                # label="90\% CI" if args.use_latex else "90% CI",
                color=args.ninety_ci_color,
            )
            ax.fill_between(
                xx,
                quant[1],
                quant[-2],
                alpha=args.fifty_ci_alpha,
                # label="50\% CI" if args.use_latex else "50% CI",
                color=args.fifty_ci_color,
            )
            ax.set_yscale(args.y_scale)
            ax.set_xscale(args.x_scale)
            if titles is not None:
                if i < len(titles):
                    ax.set_title(titles[i], fontsize=font_size)
            if x_labels is None:
                ax.set_xlabel(head, fontsize=font_size)
            else:
                if i < len(x_labels):
                    ax.set_xlabel(x_labels[i], fontsize=font_size)

            if y_labels is None:
                ax.set_ylabel(f"ppd({head})", fontsize=font_size)
            else:
                if i < len(y_labels):
                    ax.set_ylabel(y_labels[i], fontsize=font_size)

            if x_range is not None and i < len(x_range):
                ax.set_xlim(x_range[i][0], x_range[i][1])
            if y_range is not None and i < len(y_range):
                ax.set_ylim(y_range[i][0], y_range[i][1])

    for i in range(len(axes)):
        fig, ax = axes[i]
        filename = f"{args.dir}/{prefix}{headers[i]}_ppd_plot"
        if len(axes) > 1:
            filename += f"_{i}.pdf"
        else:
            filename += ".pdf"

        if args.grid:
            ax.grid(
                args.grid,
                which=args.grid_which,
                linestyle=args.grid_linestyle,
                alpha=args.grid_alpha,
            )
        ax.legend()
        fig.savefig(filename, dpi=args.dpi, bbox_inches="tight")
    plt.close("all")



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from abc import abstractmethod
from collections.abc import Callable, Sequence
from typing import Dict, Optional, Union

import equinox as eqx
from jaxtyping import Array


class VolumeTimeSensitivityInterface(eqx.Module):
    """Interface for volume time sensitivity."""

    shuffle_indices: Optional[Sequence[int]] = eqx.field(static=True, default=None)
    """The indices to shuffle the input to the model."""
    batch_size: Optional[int] = eqx.field(static=True, default=None)
    """The batch size used by :func:`jax.lax.map` in mapped functions."""
    parameter_ranges: Dict[str, Union[int, float]] = eqx.field(default=None)

    @abstractmethod
    def get_logVT(self) -> Callable[[Array], Array]:
        """Gets the log volume-time sensitivity function.

        Returns
        -------
        Callable[[Array], Array]
            A function that takes an input array of shape (n_features,) and
            returns the log volume-time sensitivity as an array of shape ().
        """
        raise NotImplementedError

    @abstractmethod
    def get_mapped_logVT(self) -> Callable[[Array], Array]:
        """Gets a mapped log volume-time sensitivity function for batch processing.

        Returns
        -------
        Callable[[Array], Array]
            A function that takes a stack of inputs as an array of shape
            (n_example, n_features) and returns an array of log volume-time
            sensitivities with shape (n_example,).
        """
        raise NotImplementedError



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import json
from collections.abc import Sequence
from typing import Dict, List, Tuple, Union

import numpy as np
import pandas as pd
from numpyro import distributions as dist
from numpyro._typing import DistributionLike

from gwkokab.vts import available as available_vts, VolumeTimeSensitivityInterface
from kokab.utils.priors import available as available_priors
from kokab.utils.regex import match_all


def read_json(json_file: str) -> Dict:
    """Read json file and return.

    Parameters
    ----------
    json_file : str
        path of the json file

    Returns
    -------
    dict
        json file content as dict

    Raises
    ------
    ValueError
        If the file is not found or if the file is not a valid json file
    """
    try:
        with open(json_file, "r") as f:
            content = json.load(f)
    except (FileNotFoundError, json.JSONDecodeError) as e:
        raise ValueError(f"Error loading configuration: {e}")
    return content


def write_json(json_file: str, content: Dict) -> None:
    """Write a dictionary to a json file.

    Parameters
    ----------
    json_file : str
        path of the json file
    content : dict
        content to write to the json file

    Raises
    ------
    ValueError
        If the file is not writable or if the content is not a valid json serializable object
    """
    try:
        with open(json_file, "w") as f:
            json.dump(content, f, indent=4)
    except (FileNotFoundError, TypeError) as e:
        raise ValueError(f"Error writing configuration: {e}")


def expand_arguments(arg: str, n: int) -> List[str]:
    """Extend the argument with a number of strings.

    .. code:: python

        >>> expand_arguments("physics", 3)
        ["physics_0", "physics_1", "physics_2"]

    Parameters
    ----------
    arg : str
        argument to extend
    n : int
        number of strings to extend

    Returns
    -------
    List[str]
        list of extended arguments
    """
    return [f"{arg}_{i}" for i in range(n)]


def get_posterior_data(
    filenames: List[str], posterior_columns: List[str]
) -> List[np.ndarray]:
    """Get the posterior data from a list of files.

    Parameters
    ----------
    filenames : List[str]
        list of filenames
    posterior_columns : List[str]
        list of posterior columns

    Returns
    -------
    List[np.ndarray]
        dictionary of posterior data

    Raises
    ------
    ValueError
        If no files are found to read posterior data
    KeyError
        If the file is missing required columns
    """
    if len(filenames) == 0:
        raise ValueError("No files found to read posterior data")
    data_list = []
    for event in filenames:
        df = pd.read_csv(event, delimiter=" ")
        missing_columns = set(posterior_columns) - set(df.columns)
        if missing_columns:
            raise KeyError(
                f"The file '{event}' is missing required columns: {missing_columns}"
            )
        data = df[posterior_columns].to_numpy()
        data_list.append(data)
    return data_list


def get_vt_samples(filename: str, columns: List[str]) -> np.ndarray:
    """Get the VT samples from a list of files.

    Parameters
    ----------
    filename : str
        list of filenames
    columns : List[str]
        list of columns

    Returns
    -------
    np.ndarray
        dictionary of VT samples

    Raises
    ------
    KeyError
        If the file is missing required columns
    """
    df = pd.read_csv(filename, delimiter=" ")
    missing_columns = set(columns) - set(df.columns)
    if missing_columns:
        raise KeyError(
            f"The file '{filename}' is missing required columns: {missing_columns}"
        )
    data = df[columns].to_numpy()
    return data


def get_processed_priors(params: List[str], priors: dict) -> dict:
    """Get the processed priors from a list of parameters.

    Parameters
    ----------
    params : List[str]
        list of parameters
    priors : dict
        dictionary of priors

    Returns
    -------
    dict
        dictionary of processed priors

    Raises
    ------
    ValueError
        if the prior value is invalid
    """
    matched_prior_params = match_all(params, priors)
    for key, value in matched_prior_params.items():
        if isinstance(value, dict):
            value_cpy = value.copy()
            dist_type = value_cpy.pop("dist")
            matched_prior_params[key] = available_priors[dist_type](
                **value_cpy, validate_args=True
            )
    for param in params:
        if param not in matched_prior_params:
            raise ValueError(f"Missing prior for {param}")
    return matched_prior_params


def check_vt_params(vt_params: List[str], parameters: List[str]) -> None:
    """Check if all the parameters in the VT are in the model.

    Parameters
    ----------
    vt_params : List[str]
        list of VT parameters
    parameters : List[str]
        list of model parameters

    Raises
    ------
    ValueError
        if the parameters in the VT do not match the parameters in the model
    """
    if set(vt_params) - set(parameters):
        raise ValueError(
            "The parameters in the VT do not match the parameters in the model. "
            f"VT_PARAMS: {vt_params}, parameters: {parameters}"
        )


def vt_json_read_and_process(
    parameters: Sequence[str], settings_path: str
) -> VolumeTimeSensitivityInterface:
    """Read and process the VT JSON file.

    Parameters
    ----------
    parameters : Sequence[str]
        list of parameters
    settings_path : str
        path to the VT settings

    Returns
    -------
    VolumeTimeSensitivityInterface
        VT object
    """
    vt_settings = read_json(settings_path)
    vt_type = vt_settings.pop("type")
    vt = available_vts[vt_type]
    if vt_type == "pdet_O3":
        return vt(parameters=parameters, **vt_settings)
    else:
        vt_path = vt_settings.pop("filename")
        return vt(parameters=parameters, filename=vt_path, **vt_settings)


def get_dist(meta_dict: dict[str, Union[str, float]]) -> DistributionLike:
    """Get the distribution from the dictionary. It expects the dictionary to have the
    key 'dist' which is the name of the distribution and the rest of the keys to be the
    parameters of the distribution.

    Example
    -------
    >>> std_normal = get_dist({"dist": "Normal", "loc": 0.0, "scale": 1.0})
    >>> std_normal.loc
    0.0
    >>> std_normal.scale
    1.0

    Parameters
    ----------
    meta_dict : dict[str, Union[str, float]]
        Dictionary containing the distribution name and its parameters

    Returns
    -------
    DistributionLike
        The distribution object
    """
    dist_name = meta_dict.pop("dist")
    return getattr(dist, dist_name)(**meta_dict)


def ppd_ranges(
    parameters: List[str], ranges: List[Tuple[str, float, float, int]]
) -> List[Tuple[float, float, int]]:
    """Convert the PPD ranges to the format required by the PPD function.

    Parameters
    ----------
    parameters : List[str]
        list of parameters
    ranges : List[Tuple[str, float, float, int]]
        list of ranges

    Returns
    -------
    List[Tuple[float, float, int]]
        list of ranges

    Raises
    ------
    ValueError
        If the parameter is not found in the list of parameters
    """
    _ranges: List[Tuple[float, float, int]] = [None] * len(parameters)
    for name, *_range in ranges:
        if name not in parameters:
            raise ValueError(f"Parameter {name} not found in {parameters}.")
        _ranges[parameters.index(name)] = (
            float(_range[0]),
            float(_range[1]),
            int(_range[2]),
        )
    return _ranges



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


#
"""
.. admonition::
    🚧 Page under construction 🚧


    Documentation for this module is under construction. You can check the source code for more information or
    run :code:`--help` on the following CLIs:

    - :code:`sage_one_powerlaw_one_peak`
    - :code:`ppd_one_powerlaw_one_peak`
"""



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from typing import Tuple

from jax import numpy as jnp
from jaxtyping import Array, ArrayLike
from quadax import cumulative_trapezoid


def beta_dist_concentrations_to_mean_variance(
    alpha: ArrayLike,
    beta: ArrayLike,
    loc: ArrayLike = 0.0,
    scale: ArrayLike = 1.0,
) -> Tuple[ArrayLike, ArrayLike]:
    r"""Let :math:`\alpha` and :math:`\beta` be the shape parameters of a beta
    distribution, :math:`a` being the location and :math:`b` being the scale. This
    function returns the mean and variance of the distribution. Then concentrations are
    given by:

    .. math::
        \mu = a+b\frac{\alpha}{\alpha + \beta}\qquad
        \sigma^2 = b^2\frac{\alpha \beta}{(\alpha + \beta)^2 (\alpha + \beta + 1)}

    Parameters
    ----------
    alpha : ArrayLike
        The shape parameter :math:`\alpha`.
    beta : ArrayLike
        The shape parameter :math:`\beta`.
    loc : ArrayLike
        The location :math:`a` of the beta distribution.
    scale : ArrayLike
        The scale :math:`b` of the beta distribution.

    Returns
    -------
    Tuple[ArrayLike, ArrayLike]
        The mean :math:`\mu` and variance :math:`\sigma^2` of the beta distribution.
    """
    sum_of_concentrations = jnp.add(alpha, beta)
    product_of_concentrations = jnp.multiply(alpha, beta)

    mean = jnp.divide(alpha, sum_of_concentrations)

    variance = jnp.add(sum_of_concentrations, 1)
    variance = jnp.multiply(jnp.square(sum_of_concentrations), variance)
    variance = jnp.divide(product_of_concentrations, variance)

    return loc + scale * mean, scale**2 * variance


def beta_dist_mean_variance_to_concentrations(
    mean: ArrayLike,
    variance: ArrayLike,
    loc: ArrayLike = 0.0,
    scale: ArrayLike = 1.0,
) -> Tuple[ArrayLike, ArrayLike]:
    r"""Let :math:`\mu` and :math:`\sigma^2` be the mean and variance of a beta
    distribution, :math:`a` being the location and :math:`b` being the scale. This
    function returns the shape parameters :math:`\alpha` and :math:`\beta` of the
    distribution. Then concentrations are given by:

    .. math::
        \alpha = -\frac{\mu-a}{b} \left(\left(\frac{\mu-a}{\sigma}\right)\left(\frac{\mu-a-b}{\sigma}\right)+1\right)\qquad
        \beta = \alpha\left(\frac{b}{\mu-a}-1\right)

    Parameters
    ----------
    mean : ArrayLike
        The mean :math:`\mu` of the beta distribution.
    variance : ArrayLike
        The variance :math:`\sigma^2` of the beta distribution.
    loc : ArrayLike
        The location :math:`a` of the beta distribution.
    scale : ArrayLike
        The scale :math:`b` of the beta distribution.

    Returns
    -------
    Tuple[ArrayLike, ArrayLike]
        The shape parameters :math:`\alpha` and :math:`\beta` of the beta distribution.
    """
    low = loc
    high = loc + scale
    mean_shifted_by_low = mean - low
    mean_shifted_by_high = mean - high
    shared_term = (mean_shifted_by_low * mean_shifted_by_high + variance) / (
        scale * variance
    )
    alpha = -mean_shifted_by_low * shared_term
    beta = mean_shifted_by_high * shared_term
    return alpha, beta


def cumtrapz(y: Array, x: Array) -> Array:
    """Calculate the cumulative trapezoidal integration of y with respect to x.

    Parameters
    ----------
    y : Array
        array to integrate
    x : Array
        array to integrate over

    Returns
    -------
    Array
        The result of the cumulative trapezoidal integration of y with respect to x.
    """
    y_cum = y
    for index in range(x.shape[-1]):
        y_cum = cumulative_trapezoid(y=y_cum, x=x[..., index], axis=index, initial=0.0)
    return y_cum



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import json
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from typing import List

from jax import numpy as jnp, random as jrd
from numpyro import distributions as dist

from gwkokab.errors import banana_error_m1_m2
from gwkokab.parameters import Parameters
from gwkokab.poisson_mean import PoissonMean
from gwkokab.population import error_magazine, PopulationFactory
from kokab.ecc_matters.common import EccentricityMattersModel
from kokab.utils import genie_parser, poisson_mean_parser
from kokab.utils.common import vt_json_read_and_process
from kokab.utils.regex import match_all


def make_parser() -> ArgumentParser:
    """Create the command line argument parser.

    This function creates the command line argument parser and returns it.

    Returns
    -------
    ArgumentParser
        the command line argument parser
    """
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser = genie_parser.get_parser(parser)
    parser.description = "Generate a population of CBCs"
    parser.epilog = "This script generates a population of CBCs"

    model_group = parser.add_argument_group("Model Options")

    model_group.add_argument(
        "--model-json",
        help="Path to the JSON file containing the model parameters",
        type=str,
        required=True,
    )

    err_group = parser.add_argument_group("Error Options")
    err_group.add_argument(
        "--err-json",
        help="Path to the JSON file containing the error parameters",
        type=str,
        required=True,
    )

    return parser


def main() -> None:
    """Main function of the script."""
    parser = make_parser()
    args = parser.parse_args()

    with open(args.model_json, "r") as f:
        model_json = json.load(f)

    with open(args.err_json, "r") as f:
        err_json = json.load(f)

    model_params_name: List[str] = [
        "alpha_m",
        "high",
        "loc",
        "log_rate",
        "low",
        "mmax",
        "mmin",
        "scale",
    ]

    model_param = match_all(model_params_name, model_json)
    err_param = match_all(
        ["scale_Mc", "scale_eta", "loc", "scale", "low", "high"], err_json
    )

    error_magazine.register(
        (Parameters.PRIMARY_MASS_SOURCE.value, Parameters.SECONDARY_MASS_SOURCE.value),
        lambda x, size, key: banana_error_m1_m2(
            x,
            size,
            key,
            scale_Mc=err_param["scale_Mc"],
            scale_eta=err_param["scale_eta"],
        ),
    )

    @error_magazine.register(Parameters.ECCENTRICITY.value)
    def ecc_error_fn(x, size, key):
        err_x = dist.TruncatedNormal(
            loc=x,
            scale=err_param["scale"],
            low=err_param["low"],
            high=err_param["high"],
        ).sample(key=key, sample_shape=(size,))
        mask = err_x < 0.0
        mask |= err_x > 1.0
        err_x = jnp.where(mask, jnp.full_like(mask, jnp.nan), err_x)
        return err_x

    model_parameters = [
        Parameters.PRIMARY_MASS_SOURCE.value,
        Parameters.SECONDARY_MASS_SOURCE.value,
        Parameters.ECCENTRICITY.value,
    ]

    nvt = vt_json_read_and_process(model_parameters, args.vt_json)
    log_selection_fn = nvt.get_mapped_logVT()

    pmean_key, factory_key = jrd.split(jrd.PRNGKey(args.seed), 2)

    pmean_kwargs = poisson_mean_parser.poisson_mean_parser(args.pmean_json)
    erate_estimator = PoissonMean(nvt, key=pmean_key, **pmean_kwargs)

    popfactory = PopulationFactory(
        model_fn=EccentricityMattersModel,
        model_params=model_param,
        parameters=model_parameters,
        log_selection_fn=log_selection_fn,
        ERate_obj=erate_estimator,
        num_realizations=args.num_realizations,
        error_size=args.error_size,
    )

    popfactory.produce(factory_key)



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from typing import Callable, Dict, List, Optional, Union

from jaxtyping import Array, ArrayLike
from numpyro._typing import DistributionLike
from numpyro.distributions.distribution import enable_validation

from gwkokab.inference import numpyro_poisson_likelihood, poisson_likelihood
from gwkokab.models import PowerlawPeak
from gwkokab.models.utils import JointDistribution
from gwkokab.parameters import Parameters
from gwkokab.poisson_mean import PoissonMean
from kokab.utils.checks import check_min_concentration_for_beta_dist
from kokab.utils.flowMC_based import flowMC_arg_parser, FlowMCBased
from kokab.utils.numpyro_based import numpyro_arg_parser, NumpyroBased
from kokab.utils.sage import Sage, sage_arg_parser


def where_fns_list(has_spin: bool) -> Optional[List[Callable[..., Array]]]:
    where_fns = []

    if has_spin:
        where_fns.append(
            lambda chi_mean,
            chi_variance,
            **kwargs: check_min_concentration_for_beta_dist(chi_mean, chi_variance)
        )

    return where_fns if len(where_fns) > 0 else None


class PowerlawPeakCore(Sage):
    def __init__(
        self,
        has_spin: bool,
        has_tilt: bool,
        has_redshift: bool,
        likelihood_fn: Callable[
            [
                Callable[..., DistributionLike],
                JointDistribution,
                Dict[str, DistributionLike],
                Dict[str, int],
                ArrayLike,
                PoissonMean,
                Optional[List[Callable[..., Array]]],
                Dict[str, Array],
            ],
            Callable,
        ],
        posterior_regex: str,
        posterior_columns: List[str],
        seed: int,
        prior_filename: str,
        selection_fn_filename: str,
        poisson_mean_filename: str,
        sampler_settings_filename: str,
        n_buckets: int,
        threshold: float,
        debug_nans: bool = False,
        profile_memory: bool = False,
        check_leaks: bool = False,
    ) -> None:
        self.has_spin = has_spin
        self.has_tilt = has_tilt
        self.has_redshift = has_redshift

        super().__init__(
            likelihood_fn=likelihood_fn,
            model=PowerlawPeak,
            posterior_regex=posterior_regex,
            posterior_columns=posterior_columns,
            seed=seed,
            prior_filename=prior_filename,
            selection_fn_filename=selection_fn_filename,
            poisson_mean_filename=poisson_mean_filename,
            sampler_settings_filename=sampler_settings_filename,
            analysis_name="one_powerlaw_one_peak",
            n_buckets=n_buckets,
            threshold=threshold,
            debug_nans=debug_nans,
            profile_memory=profile_memory,
            check_leaks=check_leaks,
            where_fns=where_fns_list(has_spin=has_spin),
        )

    @property
    def constants(self) -> Dict[str, Union[int, float, bool]]:
        return {
            "use_spin": self.has_spin,
            "use_tilt": self.has_tilt,
            "use_redshift": self.has_redshift,
        }

    @property
    def parameters(self) -> List[str]:
        names = [
            Parameters.PRIMARY_MASS_SOURCE.value,
            Parameters.SECONDARY_MASS_SOURCE.value,
        ]
        if self.has_spin:
            names.append(Parameters.PRIMARY_SPIN_MAGNITUDE.value)
            names.append(Parameters.SECONDARY_SPIN_MAGNITUDE.value)
        if self.has_tilt:
            names.extend([Parameters.COS_TILT_1.value, Parameters.COS_TILT_2.value])
        if self.has_redshift:
            names.append(Parameters.REDSHIFT.value)
        return names

    @property
    def model_parameters(self) -> List[str]:
        model_parameters = [
            "alpha",
            "beta",
            "delta",
            "lambda_peak",
            "loc",
            "log_rate",
            "mmax",
            "mmin",
            "scale",
        ]

        if self.has_spin:
            model_parameters.extend(["chi_mean", "chi_variance"])

        if self.has_tilt:
            model_parameters.extend(["cos_tilt_zeta", "cos_tilt_scale"])

        if self.has_redshift:
            model_parameters.extend(["kappa", "z_max"])

        return model_parameters


class PowerlawPeakFSage(PowerlawPeakCore, FlowMCBased):
    pass


class PowerlawPeakNSage(PowerlawPeakCore, NumpyroBased):
    pass


def model_arg_parser(parser: ArgumentParser) -> ArgumentParser:
    model_group = parser.add_argument_group("Model Options")
    model_group.add_argument(
        "--add-spin",
        action="store_true",
        help="Include beta spin parameters in the model.",
    )
    model_group.add_argument(
        "--add-tilt",
        action="store_true",
        help="Include tilt parameters in the model.",
    )
    model_group.add_argument(
        "--add-redshift",
        action="store_true",
        help="Include redshift parameter in the model",
    )
    return parser


def f_main() -> None:
    enable_validation()

    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser = model_arg_parser(parser)
    parser = sage_arg_parser(parser)
    parser = flowMC_arg_parser(parser)

    args = parser.parse_args()

    PowerlawPeakFSage(
        has_spin=args.add_spin,
        has_tilt=args.add_tilt,
        has_redshift=args.add_redshift,
        likelihood_fn=poisson_likelihood,
        posterior_regex=args.posterior_regex,
        posterior_columns=args.posterior_columns,
        seed=args.seed,
        prior_filename=args.prior_json,
        selection_fn_filename=args.vt_json,
        poisson_mean_filename=args.pmean_json,
        sampler_settings_filename=args.sampler_config,
        n_buckets=args.n_buckets,
        threshold=args.threshold,
        debug_nans=args.debug_nans,
        profile_memory=args.profile_memory,
        check_leaks=args.check_leaks,
    ).run()


def n_main() -> None:
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser = model_arg_parser(parser)
    parser = sage_arg_parser(parser)
    parser = numpyro_arg_parser(parser)

    args = parser.parse_args()

    PowerlawPeakNSage(
        has_spin=args.add_spin,
        has_tilt=args.add_tilt,
        has_redshift=args.add_redshift,
        likelihood_fn=numpyro_poisson_likelihood,
        posterior_regex=args.posterior_regex,
        posterior_columns=args.posterior_columns,
        seed=args.seed,
        prior_filename=args.prior_json,
        selection_fn_filename=args.vt_json,
        poisson_mean_filename=args.pmean_json,
        sampler_settings_filename=args.sampler_config,
        n_buckets=args.n_buckets,
        threshold=args.threshold,
        debug_nans=args.debug_nans,
        profile_memory=args.profile_memory,
        check_leaks=args.check_leaks,
    ).run()



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from ._models import (
    BetaFromMeanVar as BetaFromMeanVar,
    GaussianSpinModel as GaussianSpinModel,
    IndependentSpinOrientationGaussianIsotropic as IndependentSpinOrientationGaussianIsotropic,
)



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0


from collections import namedtuple

import jax
import jax.numpy as jnp
import pytest
from jax import jit, vmap

from gwkokab.models import constraints


_a = jnp.asarray


class T(namedtuple("TestCase", ["constraint_cls", "params", "kwargs"])):
    pass


SINGLETON_CONSTRAINTS = {
    "decreasing_vector": constraints.decreasing_vector,
    "increasing_vector": constraints.increasing_vector,
    "positive_decreasing_vector": constraints.positive_decreasing_vector,
    "positive_increasing_vector": constraints.positive_increasing_vector,
    "strictly_decreasing_vector": constraints.strictly_decreasing_vector,
    "strictly_increasing_vector": constraints.strictly_increasing_vector,
}
PARAMETRIZED_CONSTRAINTS = {
    "mass_sandwich": T(constraints.mass_sandwich, (_a(10.0), _a(30.0)), dict()),
    "mass_ratio_mass_sandwich": T(
        constraints.mass_ratio_mass_sandwich, (_a(10.0), _a(30.0)), dict()
    ),
    "all_constraint": T(
        constraints.all_constraint,
        (),
        {
            "constraints": [
                constraints.decreasing_vector,
                constraints.positive_decreasing_vector,
            ],
            "event_slices": [0, 1],
        },
    ),
    "any_constraint": T(
        constraints.any_constraint,
        (),
        {
            "constraints": [
                constraints.decreasing_vector,
                constraints.positive_decreasing_vector,
            ]
        },
    ),
}


@pytest.mark.parametrize(
    "constraint", SINGLETON_CONSTRAINTS.values(), ids=SINGLETON_CONSTRAINTS.keys()
)
def test_singleton_constraint_pytree(constraint):
    # test that singleton constraints objects can be used as pytrees
    def in_cst(constraint, x):
        return x**2

    def out_cst(constraint, x):
        return constraint

    jitted_in_cst = jit(in_cst)
    jitted_out_cst = jit(out_cst)

    assert jitted_in_cst(constraint, 1.0) == 1.0
    assert jitted_out_cst(constraint, 1.0) == constraint

    assert jnp.allclose(
        vmap(in_cst, in_axes=(None, 0), out_axes=0)(constraint, jnp.ones(3)),
        jnp.ones(3),
    )

    assert (
        vmap(out_cst, in_axes=(None, 0), out_axes=None)(constraint, jnp.ones(3))
        is constraint
    )


@pytest.mark.parametrize(
    "cls, cst_args, cst_kwargs",
    PARAMETRIZED_CONSTRAINTS.values(),
    ids=PARAMETRIZED_CONSTRAINTS.keys(),
)
def test_parametrized_constraint_pytree(cls, cst_args, cst_kwargs):
    constraint = cls(*cst_args, **cst_kwargs)

    # test that singleton constraints objects can be used as pytrees
    def in_cst(constraint, x):
        return x**2

    def out_cst(constraint, x):
        return constraint

    jitted_in_cst = jit(in_cst)
    jitted_out_cst = jit(out_cst)

    assert jitted_in_cst(constraint, 1.0) == 1.0
    assert jitted_out_cst(constraint, 1.0) == constraint

    assert jnp.allclose(
        vmap(in_cst, in_axes=(None, 0), out_axes=0)(constraint, jnp.ones(3)),
        jnp.ones(3),
    )

    assert (
        vmap(out_cst, in_axes=(None, 0), out_axes=None)(constraint, jnp.ones(3))
        == constraint
    )

    if len(cst_args) > 0:
        # test creating and manipulating vmapped constraints
        vmapped_cst_args = jax.tree.map(lambda x: x[None], cst_args)

        vmapped_csts = jit(vmap(lambda args: cls(*args, **cst_kwargs), in_axes=(0,)))(
            vmapped_cst_args
        )
        assert vmap(lambda x: x == constraint, in_axes=0)(vmapped_csts).all()

        twice_vmapped_cst_args = jax.tree.map(lambda x: x[None], vmapped_cst_args)

        vmapped_csts = jit(
            vmap(
                vmap(lambda args: cls(*args, **cst_kwargs), in_axes=(0,)),
                in_axes=(0,),
            ),
        )(twice_vmapped_cst_args)
        assert vmap(vmap(lambda x: x == constraint, in_axes=0), in_axes=0)(
            vmapped_csts
        ).all()


@pytest.mark.parametrize(
    "constraint", SINGLETON_CONSTRAINTS.values(), ids=SINGLETON_CONSTRAINTS.keys()
)
def test_singleton_constraint_eq(constraint):
    assert constraint == constraint
    assert constraint != 1

    # check that equality checks are robust to constraints parametrized
    # by abstract values
    @jit
    def check_constraints(c1, c2):
        return c1 == c2

    assert check_constraints(constraint, constraint)


@pytest.mark.parametrize(
    "cls, cst_args, cst_kwargs",
    PARAMETRIZED_CONSTRAINTS.values(),
    ids=PARAMETRIZED_CONSTRAINTS.keys(),
)
def test_parametrized_constraint_eq(cls, cst_args, cst_kwargs):
    constraint = cls(*cst_args, **cst_kwargs)
    constraint2 = cls(*cst_args, **cst_kwargs)
    assert constraint == constraint2
    assert constraint != 1

    # check that equality checks are robust to constraints parametrized
    # by abstract values
    @jit
    def check_constraints(c1, c2):
        return c1 == c2

    assert check_constraints(constraint, constraint2)



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0

from typing import Optional

import equinox as eqx
import jax.numpy as jnp
import quadax
from jaxtyping import ArrayLike

from gwkokab.constants import SPEED_OF_LIGHT


class Cosmology(eqx.Module):
    """Flat ΛCDM cosmology with Mpc-based units for distances and comoving volumes."""

    _Ho: ArrayLike
    _OmegaKappa: ArrayLike
    _OmegaLambda: ArrayLike
    _OmegaMatter: ArrayLike
    _OmegaRadiation: ArrayLike
    _z: ArrayLike
    _Dc: ArrayLike
    Vc: ArrayLike

    def __init__(
        self,
        Ho: ArrayLike,
        omega_matter: ArrayLike,
        omega_radiation: ArrayLike,
        omega_lambda: ArrayLike,
        max_z: float = 4.0,
        dz: float = 1e-3,
    ) -> None:
        """Initialize the cosmology with given parameters.

        Parameters
        ----------
        Ho : ArrayLike
            Hubble constant in m/s/Mpc.
        omega_matter : ArrayLike
            Matter density parameter (Ω_m).
        omega_radiation : ArrayLike
            Radiation density parameter (Ω_r).
        omega_lambda : ArrayLike
            Dark energy density parameter (Ω_Λ).
        max_z : float, optional
            Maximum redshift to consider, by default 4.0
        dz : float, optional
            Redshift bin width, by default DEFAULT_DZ
        """
        self._Ho = Ho  # Hubble constant in m/s/Mpc
        self._OmegaMatter = omega_matter
        self._OmegaRadiation = omega_radiation
        self._OmegaLambda = omega_lambda
        self._OmegaKappa = 1.0 - (
            self._OmegaMatter + self._OmegaRadiation + self._OmegaLambda
        )

        assert jnp.isclose(self._OmegaKappa, 0.0, atol=1e-10), (
            "Only flat cosmologies are supported (Ω_k ≈ 0)."
        )

        self._z = jnp.arange(0.0, max_z + dz, dz)
        self._Dc = quadax.cumulative_trapezoid(
            self.dDcdz(self._z), x=self._z, initial=0.0, axis=0
        )
        self.Vc = quadax.cumulative_trapezoid(
            self.dVcdz(self._z, self._Dc), x=self._z, initial=0.0, axis=0
        )

    # --------- Core functions ---------
    def z_to_E(self, z: ArrayLike) -> ArrayLike:
        zp1 = 1.0 + z
        return jnp.sqrt(
            self._OmegaLambda
            + self._OmegaKappa * zp1**2
            + self._OmegaMatter * zp1**3
            + self._OmegaRadiation * zp1**4
        )

    def dDcdz(self, z: ArrayLike) -> ArrayLike:
        return (SPEED_OF_LIGHT / self._Ho) / self.z_to_E(z)

    def logdVcdz(self, z: ArrayLike, Dc: Optional[ArrayLike] = None) -> ArrayLike:
        if Dc is None:
            Dc = self.z_to_Dc(z)
        return (
            jnp.log(4.0 * jnp.pi)  # Conversion of steradians
            - 9.0 * jnp.log(10.0)  # Conversion from Mpc^3 to Gpc^3
            + 2 * jnp.log(Dc)
            + jnp.log(self.dDcdz(z))
        )

    def dVcdz(self, z: ArrayLike, Dc: Optional[ArrayLike] = None) -> ArrayLike:
        return jnp.exp(self.logdVcdz(z, Dc=Dc))

    # --------- Interpolators ---------
    def z_to_Dc(self, z: ArrayLike) -> ArrayLike:
        """Fast JAX-safe interpolation of comoving distance."""
        return jnp.interp(z, self._z, self._Dc)

    def z_to_DL(self, z: ArrayLike) -> ArrayLike:
        """Luminosity distance in Mpc."""
        return jnp.interp(z, self._z, self.DL)

    def DL_to_z(self, DL: ArrayLike) -> ArrayLike:
        """Approximate inversion DL -> z using precomputed grid."""
        return jnp.interp(DL, self.DL, self._z)

    # --------- Properties ---------
    @property
    def DL(self) -> ArrayLike:
        return self._Dc * (1.0 + self._z)

    @property
    def z(self) -> ArrayLike:
        return self._z

    @property
    def Dc(self) -> ArrayLike:
        return self._Dc



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from numpyro.distributions import TransformedDistribution

from ..constraints import transform_constraint


class ExtendedSupportTransformedDistribution(TransformedDistribution):
    @property
    def support(self):
        return transform_constraint(self.base_dist.support, self.transforms)



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from typing import List, Optional

import jax
from jax import lax, numpy as jnp
from jaxtyping import Array, ArrayLike
from numpyro.distributions import constraints, Distribution
from numpyro.distributions.util import categorical, is_prng_key, validate_sample


class ScaledMixture(Distribution):
    r"""A finite mixture of component distributions from different families. This is a
    generalization of :class:`~numpyro.distributions.Mixture` where the component
    distributions are scaled by a set of rates.

    **Example**

    .. code::

       >>> import jax
       >>> import jax.random as jrd
       >>> import numpyro.distributions as dist
       >>> from gwkokab.models.utils import ScaledMixture
       >>> log_scales = jrd.uniform(jrd.PRNGKey(42), (3,), minval=0, maxval=5)
       >>> component_dists = [
       ...     dist.Normal(loc=0.0, scale=1.0),
       ...     dist.Normal(loc=-0.5, scale=0.3),
       ...     dist.Normal(loc=0.6, scale=1.2),
       ... ]
       >>> mixture = ScaledMixture(log_scales, component_dists)
       >>> mixture.sample(jax.random.PRNGKey(42)).shape
       ()
    """

    arg_constraints = {
        "log_scales": constraints.real_vector,
    }
    pytree_data_fields = ("_component_distributions", "_support", "log_scales")
    pytree_aux_fields = ("_mixture_size",)

    def __init__(
        self,
        log_scales: Array,
        component_distributions: List[Distribution],
        *,
        support: Optional[constraints.Constraint] = None,
        validate_args: Optional[bool] = None,
    ):
        self.log_scales = log_scales
        try:
            component_distributions = list(component_distributions)
        except TypeError:
            raise ValueError(
                "The 'component_distributions' argument must be a list of Distribution objects"
            )
        self._mixture_size = log_scales.shape[-1]
        for d in component_distributions:
            if not isinstance(d, Distribution):
                raise ValueError(
                    "All elements of 'component_distributions' must be instances of "
                    "numpyro.distributions.Distribution subclasses"
                )
        if len(component_distributions) != self.mixture_size:
            raise ValueError(
                "The number of elements in 'component_distributions' must match the mixture size; "
                f"expected {self._mixture_size}, got {len(component_distributions)}"
            )

        # TODO: It would be good to check that the support of all the component
        # distributions match, but for now we just check the type, since __eq__
        # isn't consistently implemented for all support types.
        self._support = support
        if support is None:
            support_type = type(component_distributions[0].support)
            if any(
                type(d.support) is not support_type for d in component_distributions[1:]
            ):
                raise ValueError(
                    "All component distributions must have the same support."
                )
        else:
            assert isinstance(support, constraints.Constraint), (
                "support must be a Constraint object"
            )

        self._component_distributions = component_distributions

        batch_shape = lax.broadcast_shapes(
            *(d.batch_shape for d in component_distributions)
        )
        event_shape = component_distributions[0].event_shape
        for d in component_distributions[1:]:
            if d.event_shape != event_shape:
                raise ValueError(
                    "All component distributions must have the same event shape"
                )

        super().__init__(
            batch_shape=batch_shape,
            event_shape=event_shape,
            validate_args=validate_args,
        )

    @property
    def component_distributions(self):
        """The list of component distributions in the mixture.

        :return: The list of component distributions
        :rtype: list[Distribution]
        """
        return self._component_distributions

    @constraints.dependent_property
    def support(self):
        if self._support is not None:
            return self._support
        return self.component_distributions[0].support

    @property
    def is_discrete(self):
        return self.component_distributions[0].is_discrete

    @property
    def component_mean(self):
        return jnp.stack(
            [d.mean for d in self.component_distributions], axis=self.mixture_dim
        )

    @property
    def component_variance(self):
        return jnp.stack(
            [d.variance for d in self.component_distributions], axis=self.mixture_dim
        )

    def component_cdf(self, samples):
        return jnp.stack(
            [d.cdf(samples) for d in self.component_distributions],
            axis=self.mixture_dim,
        )

    def component_sample(self, key, sample_shape=()):
        keys = jax.random.split(key, self.mixture_size)
        samples = []
        for k, d in zip(keys, self.component_distributions):
            samples.append(d.expand(sample_shape + self.batch_shape).sample(k))
        return jnp.stack(samples, axis=self.mixture_dim)

    def component_log_probs(self, value: ArrayLike) -> ArrayLike:
        # modified implementation of numpyro.distributions.MixtureGeneral.component_log_probs
        component_log_probs = []
        for d in self.component_distributions:
            log_prob = d.log_prob(value)
            if (self._support is not None) and (not d._validate_args):
                mask = d.support(value)
                log_prob = jnp.where(mask, log_prob, -jnp.inf)
            component_log_probs.append(log_prob)
        component_log_probs = jnp.stack(component_log_probs, axis=-1)
        return self.log_scales + component_log_probs

    @property
    def mixture_size(self):
        """The number of components in the mixture."""
        return self._mixture_size

    @property
    def mixture_dim(self):
        return -self.event_dim - 1

    @property
    def mean(self):
        probs = jnp.exp(self.log_scales)
        probs = probs.reshape(probs.shape + (1,) * self.event_dim)
        weighted_component_means = probs * self.component_mean
        return jnp.sum(weighted_component_means, axis=self.mixture_dim)

    @property
    def variance(self):
        # TODO(Qazalbash): Check the correctness
        probs = jnp.exp(self.log_scales)
        probs = probs.reshape(probs.shape + (1,) * self.event_dim)
        mean_cond_var = jnp.sum(probs * self.component_variance, axis=self.mixture_dim)
        sq_deviation = (
            self.component_mean - jnp.expand_dims(self.mean, axis=self.mixture_dim)
        ) ** 2
        var_cond_mean = jnp.sum(probs * sq_deviation, axis=self.mixture_dim)
        return mean_cond_var + var_cond_mean

    def cdf(self, samples):
        """The cumulative distribution function.

        :param value: samples from this distribution.
        :return: output of the cumulative distribution function evaluated at
            `value`.
        :raises: NotImplementedError if the component distribution does not
            implement the cdf method.
        """
        cdf_components = self.component_cdf(samples)
        return jnp.sum(cdf_components * jnp.exp(self.log_scales), axis=-1)

    def sample_with_intermediates(self, key, sample_shape=()):
        """A version of ``sample`` that also returns the sampled component indices.

        :param jax.random.PRNGKey key: the rng_key key to be used for the distribution.
        :param tuple sample_shape: the sample shape for the distribution.
        :return: A 2-element tuple with the samples from the distribution, and the
            indices of the sampled components.
        :rtype: tuple
        """
        assert is_prng_key(key)
        key_comp, key_ind = jax.random.split(key)
        samples = self.component_sample(key_comp, sample_shape=sample_shape)

        # Sample selection indices from the categorical (shape will be sample_shape)
        indices: Array = categorical(
            key_ind,
            jax.nn.softmax(self.log_scales, axis=-1),
            shape=sample_shape + self.batch_shape,
        )
        n_expand = self.event_dim + 1
        indices_expanded = indices.reshape(indices.shape + (1,) * n_expand)

        # Select samples according to indices samples from categorical
        samples_selected = jnp.take_along_axis(
            samples, indices=indices_expanded, axis=self.mixture_dim
        )

        # Final sample shape (*sample_shape, *batch_shape, *event_shape)
        return jnp.squeeze(samples_selected, axis=self.mixture_dim), [indices]

    def sample(self, key, sample_shape=()):
        return self.sample_with_intermediates(key=key, sample_shape=sample_shape)[0]

    @validate_sample
    def log_prob(self, value, intermediates=None):
        del intermediates
        sum_log_probs = self.component_log_probs(value)
        safe_sum_log_probs = jnp.where(
            jnp.isneginf(sum_log_probs), -jnp.inf, sum_log_probs
        )
        return jax.nn.logsumexp(
            safe_sum_log_probs,
            where=~jnp.isneginf(sum_log_probs),
            axis=-1,
        )



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import warnings
from typing import Dict, Optional, Tuple, TypeVar

import jax
from jaxtyping import Array
from loguru import logger


_KT = TypeVar("_KT")
_VT = TypeVar("_VT")


def fetch_first_matching_value(dictionary: Dict[_KT, _VT], *keys: _KT) -> Optional[_VT]:
    """Get the first value in the dictionary that matches one of the keys.

    Parameters
    ----------
    dictionary : Dict[_KT, _VT]
        The dictionary to search.
    keys : _KT
        The keys to search for.

    Returns
    -------
    Optional[_VT]
        The value of the first key that is found in the dictionary, or None if no key is
        found.
    """
    for key in keys:
        if key in dictionary:
            return dictionary[key]
    return None


def error_if(cond: bool, err: type[Exception] = ValueError, msg: str = "") -> None:
    """Raise an error if condition is met.

    Reference: utils of `interpax <https://github.com/f0uriest/interpax>`_.

    Parameters
    ----------
    cond : bool
        The condition to check.
    err : Exception, optional
        The error to raise, by default ValueError
    msg : str, optional
        The message to include with the error, by default ""

    Raises
    ------
    err
        The error raised if the condition is met.
    """
    if cond:
        logger.error(msg)
        raise err(msg)


def warn_if(cond: bool, err: type[Warning] = UserWarning, msg: str = "") -> None:
    """Raise a warning if condition is met.

    Reference: utils of `interpax <https://github.com/f0uriest/interpax>`_.

    Parameters
    ----------
    cond : bool
        The condition to check.
    err : Warning, optional
        The warning to raise, by default UserWarning
    msg : str, optional
        The message to include with the warning, by default ""
    """
    if cond:
        logger.warning(msg)
        warnings.warn(msg, err)


def batch_and_remainder(x: Array, batch_size: int) -> Tuple[Array, Array]:
    """Calculated batch and remainder of an array given a batch size.

    Copied from JAX codebase.

    Parameters
    ----------
    x : Array
        Array of interest
    batch_size : int
        batch size

    Returns
    -------
    Tuple[Array, Array]
        batched array and remainder
    """
    leaves, treedef = jax.tree_util.tree_flatten(x)

    scan_leaves = []
    remainder_leaves = []

    for leaf in leaves:
        num_batches, _ = divmod(leaf.shape[0], batch_size)
        total_batch_elems = num_batches * batch_size
        scan_leaves.append(
            leaf[:total_batch_elems].reshape(num_batches, batch_size, *leaf.shape[1:])
        )
        remainder_leaves.append(leaf[total_batch_elems:])

    scan_tree = treedef.unflatten(scan_leaves)
    remainder_tree = treedef.unflatten(remainder_leaves)
    return scan_tree, remainder_tree



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import math
from collections import namedtuple

import jax
import jax.numpy as jnp
import pytest
from jax import jacfwd, jit, random, vmap
from numpyro.distributions.transforms import biject_to

from gwkokab.models import constraints
from gwkokab.models.transformations import (
    ComponentMassesAndRedshiftToDetectedMassAndRedshift,
    ComponentMassesToChirpMassAndDelta,
    ComponentMassesToChirpMassAndSymmetricMassRatio,
    ComponentMassesToMassRatioAndSecondaryMass,
    ComponentMassesToPrimaryMassAndMassRatio,
    ComponentMassesToTotalMassAndMassRatio,
    DeltaToSymmetricMassRatio,
    PrimaryMassAndMassRatioToComponentMassesTransform,
    SourceMassAndRedshiftToDetectedMassAndRedshift,
)


class T(namedtuple("TestCase", ["transform_cls", "params", "kwargs"])):
    pass


TRANSFORMS = {
    "component_masses_and_redshift_to_detected_mass_and_redshift": T(
        ComponentMassesAndRedshiftToDetectedMassAndRedshift, (), dict()
    ),
    "component_masses_to_chirp_mass_and_delta": T(
        ComponentMassesToChirpMassAndDelta, (), dict()
    ),
    "component_masses_to_chirp_mass_and_symmetric_mass_ratio": T(
        ComponentMassesToChirpMassAndSymmetricMassRatio, (), dict()
    ),
    "component_masses_to_mass_ratio_and_secondary_mass": T(
        ComponentMassesToMassRatioAndSecondaryMass, (), dict()
    ),
    "component_masses_to_primary_mass_and_mass_ratio": T(
        ComponentMassesToPrimaryMassAndMassRatio, (), dict()
    ),
    "component_masses_to_total_mass_and_mass_ratio": T(
        ComponentMassesToTotalMassAndMassRatio, (), dict()
    ),
    "delta_to_symmetric_mass_ratio": T(DeltaToSymmetricMassRatio, (), dict()),
    "primary_mass_and_mass_ratio_to_component_masses_transform": T(
        PrimaryMassAndMassRatioToComponentMassesTransform, (), dict()
    ),
    "source_mass_and_redshift_to_detected_mass_and_redshift": T(
        SourceMassAndRedshiftToDetectedMassAndRedshift, (), dict()
    ),
}


@pytest.mark.parametrize(
    "cls, transform_args, transform_kwargs",
    TRANSFORMS.values(),
    ids=TRANSFORMS.keys(),
)
def test_parametrized_transform_pytree(cls, transform_args, transform_kwargs):
    transform = cls(*transform_args, **transform_kwargs)

    # test that singleton transforms objects can be used as pytrees
    def in_t(transform, x):
        return x**2

    def out_t(transform, x):
        return transform

    jitted_in_t = jit(in_t)
    jitted_out_t = jit(out_t)

    assert jitted_in_t(transform, 1.0) == 1.0
    assert jitted_out_t(transform, 1.0) == transform

    assert jitted_out_t(transform.inv, 1.0) == transform.inv

    assert jnp.allclose(
        vmap(in_t, in_axes=(None, 0), out_axes=0)(transform, jnp.ones(3)),
        jnp.ones(3),
    )

    assert (
        vmap(out_t, in_axes=(None, 0), out_axes=None)(transform, jnp.ones(3))
        == transform
    )

    if len(transform_args) > 0:
        # test creating and manipulating vmapped constraints
        # this test assumes jittable args, and non-jittable kwargs, which is
        # not suited for all transforms, see InverseAutoregressiveTransform.
        # TODO: split among jittable and non-jittable args/kwargs instead.
        vmapped_transform_args = jax.tree.map(lambda x: x[None], transform_args)

        vmapped_transform = jit(
            vmap(lambda args: cls(*args, **transform_kwargs), in_axes=(0,))
        )(vmapped_transform_args)
        assert vmap(lambda x: x == transform, in_axes=0)(vmapped_transform).all()

        twice_vmapped_transform_args = jax.tree.map(
            lambda x: x[None], vmapped_transform_args
        )

        vmapped_transform = jit(
            vmap(
                vmap(lambda args: cls(*args, **transform_kwargs), in_axes=(0,)),
                in_axes=(0,),
            )
        )(twice_vmapped_transform_args)
        assert vmap(vmap(lambda x: x == transform, in_axes=0), in_axes=0)(
            vmapped_transform
        ).all()


@pytest.mark.parametrize(
    "cls, transform_args, transform_kwargs",
    TRANSFORMS.values(),
    ids=TRANSFORMS.keys(),
)
def test_parametrized_transform_eq(cls, transform_args, transform_kwargs):
    transform = cls(*transform_args, **transform_kwargs)
    transform2 = cls(*transform_args, **transform_kwargs)
    assert transform == transform2
    assert transform != 1.0

    # check that equality checks are robust to transforms parametrized
    # by abstract values
    @jit
    def check_transforms(t1, t2):
        return t1 == t2

    assert check_transforms(transform, transform2)


@pytest.mark.parametrize(
    "transform, shape",
    [
        (ComponentMassesAndRedshiftToDetectedMassAndRedshift(), (3,)),
        (ComponentMassesToChirpMassAndDelta(), (2,)),
        (ComponentMassesToChirpMassAndSymmetricMassRatio(), (2,)),
        (ComponentMassesToMassRatioAndSecondaryMass(), (2,)),
        (ComponentMassesToPrimaryMassAndMassRatio(), (2,)),
        (ComponentMassesToTotalMassAndMassRatio(), (2,)),
        (DeltaToSymmetricMassRatio(), ()),
        (PrimaryMassAndMassRatioToComponentMassesTransform(), (2,)),
        (SourceMassAndRedshiftToDetectedMassAndRedshift(), (2,)),
    ],
)
def test_bijective_transforms(transform, shape):
    if isinstance(transform, type):
        pytest.skip()
    # Get a sample from the support of the distribution.
    batch_shape = (13,)
    unconstrained = random.normal(random.key(17), batch_shape + shape)
    x1 = biject_to(transform.domain)(unconstrained)

    # Transform forward and backward, checking shapes, values, and Jacobian shape.
    y = transform(x1)
    assert y.shape == transform.forward_shape(x1.shape)

    x2 = transform.inv(y)
    assert x2.shape == transform.inverse_shape(y.shape)
    # Some transforms are a bit less stable; we give them larger tolerances.
    atol = 1e-6
    assert jnp.allclose(x1, x2, atol=atol, equal_nan=True)

    log_abs_det_jacobian = transform.log_abs_det_jacobian(x1, y)
    assert log_abs_det_jacobian.shape == batch_shape

    # Also check the Jacobian numerically for transforms with the same input and output
    # size, unless they are explicitly excluded. E.g., the upper triangular of the
    # CholeskyTransform is zero, giving rise to a singular Jacobian.
    size_x = int(x1.size / math.prod(batch_shape))
    size_y = int(y.size / math.prod(batch_shape))
    if size_x == size_y:
        jac = (
            vmap(jacfwd(transform))(x1)
            .reshape((-1,) + x1.shape[len(batch_shape) :])
            .reshape(batch_shape + (size_y, size_x))
        )
        slogdet = jnp.linalg.slogdet(jac)
        assert jnp.allclose(log_abs_det_jacobian, slogdet.logabsdet, atol=atol)


@pytest.mark.parametrize(
    "constraint, shape",
    [
        (constraints.positive_decreasing_vector, (5,)),
        (constraints.decreasing_vector, (5,)),
        (constraints.strictly_decreasing_vector, (5,)),
        (constraints.positive_increasing_vector, (5,)),
        (constraints.increasing_vector, (5,)),
        (constraints.strictly_increasing_vector, (5,)),
        (constraints.mass_sandwich(10.0, 50.0), (2,)),
        (constraints.mass_ratio_mass_sandwich(10.0, 50.0), (2,)),
    ],
    ids=str,
)
def test_biject_to(constraint, shape):
    batch_shape = (13, 19)
    unconstrained = random.normal(random.key(93), batch_shape + shape)
    constrained = biject_to(constraint)(unconstrained)
    passed = constraint.check(constrained)
    expected_shape = constrained.shape[: constrained.ndim - constraint.event_dim]
    assert passed.shape == expected_shape
    assert jnp.all(passed)



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from functools import partial
from typing import Dict, List, Tuple, Union

import h5py
import jax
import numpy as np
from jax import numpy as jnp
from jaxtyping import Array, ArrayLike
from numpyro._typing import DistributionLike

from gwkokab.utils.tools import error_if


def wipe_log_rate(
    nf_samples: Array,
    nf_samples_mapping: Dict[str, int],
    constants: Dict[str, Union[int, float]],
) -> Tuple[Array, Dict[str, Union[int, float]]]:
    """Set the log rate parameters to zero and remove them from the samples.

    Parameters
    ----------
    nf_samples : Array
        Normalizing flow samples.
    nf_samples_mapping : Dict[str, int]
        Mapping of the normalizing flow samples.
    constants : Dict[str, Union[int, float]]
        Constants.

    Returns
    -------
    Tuple[Array, Dict[str, Union[int, float]]]
        The normalizing flow samples and the updated constants.
    """
    for key in list(nf_samples_mapping.keys()).copy():
        if key.startswith("log_rate"):
            index = nf_samples_mapping.pop(key)
            constants[key] = 0.0
            nf_samples = np.delete(nf_samples, index, axis=-1)
            for key, value in nf_samples_mapping.items():
                if value > index:
                    nf_samples_mapping[key] -= 1
    return nf_samples, constants


def compute_probs(
    params: Array,
    xx_mesh: Array,
    model: DistributionLike,
    constants: Dict[str, ArrayLike],
    nf_samples_mapping: Dict[str, int],
    batch_size: int = 1000,
) -> Array:
    """Compute the probability density function of a model.

    Parameters
    ----------
    params : Array
        A callable that computes the log-probability density function of the model.
    xx_mesh : Array
        A list of tuples `(start, end, num_points)` for each parameter, defining the
        grid over which to compute the PPD.
    model : DistributionLike
        The PPD of the model as a multidimensional array corresponding to the parameter
        grid.
    constants : Dict[str, ArrayLike]
        A dictionary of constants for the model.
    nf_samples_mapping : Dict[str, int]
        A dictionary mapping the normalizing flow samples to the model parameters.
    batch_size : int
        The batch size for the computation, defaults to 1000.

    Returns
    -------
    Array
        The probability density function of the model.
    """

    logpdf = model(
        **constants,
        **{k: params[v] for k, v in nf_samples_mapping.items()},
        validate_args=True,
    ).log_prob

    def _prob(x: Array) -> Array:
        x_expanded = jnp.expand_dims(x, axis=-2)
        prob = jnp.exp(logpdf(x_expanded))
        return prob

    prob_vec = jax.lax.map(_prob, xx_mesh, batch_size=batch_size)

    return prob_vec


def _compute_marginal_probs(
    probs_array: Array,
    axis: int,
    domain: List[Tuple[float, float, int]],
) -> Array:
    """Compute the marginal probabilities of a model.

    The function computes the marginal probabilities of a model by summing over the
    specified axis.

    Parameters
    ----------
    probs_array : Array
        The probabilities of the model.
    axis : int
        The axis along which to compute the marginal probabilities.
    domain : List[Tuple[float, float, int]]
        The domain of the axis.

    Returns
    -------
    Array
        The marginal probabilities of the model.
    """
    assert axis < probs_array.ndim, "Axis must be less than the number of dimensions."
    j = 0
    marginal_density = probs_array
    for i, (start, end, num_points) in enumerate(domain):
        if i == axis:
            continue
        num_points = int(num_points)
        marginal_density = jnp.trapezoid(
            y=marginal_density,
            x=jnp.linspace(start, end, num_points),
            axis=i - j,
        )
        j += 1

    return marginal_density


def get_all_marginals(
    probs: Array,
    domains: List[Tuple[float, float, int]],
) -> List[Array]:
    """Compute marginal probabilities for all axes.

    Parameters
    ----------
    probs : Array
        The probability array.
    domains : List[Tuple[float, float, int]]
        List of domains for each axis.

    Returns
    -------
    List[Array]
        List of marginal probability arrays, one for each axis.
    """
    return [_compute_marginal_probs(probs, axis, domains) for axis in range(probs.ndim)]


def save_probs(
    ppd_array: Array,
    marginal_probs: List[Array],
    filename: str,
    domains: List[Tuple[float, float, int]],
    headers: List[str],
) -> None:
    """Save the PPD and marginal probabilities to a file.

    Parameters
    ----------
    ppd_array : Array
        The ppd array.
    marginal_probs : List[Array]
        List of marginal probabilities.
    filename : str
        The name of the file to save the PPD and marginal probabilities.
    domains : List[Tuple[float, float, int]]
        List of domains for each axis
    headers : List[str]
        List of headers for the PPD and marginal probabilities
    """
    error_if(
        ppd_array.ndim != len(domains),
        ValueError,
        "Number of ranges must match the number of dimensions of the PPD array.",
    )
    error_if(
        ppd_array.ndim != len(headers),
        ValueError,
        "Number of headers must match the number of dimensions of the PPD array.",
    )

    with h5py.File(filename, "w") as f:
        f.create_dataset("domains", data=np.array(domains))
        f.create_dataset("headers", data=np.array(headers, dtype="S"))
        f.create_dataset("ppd", data=ppd_array)
        marginal_probs_group = f.create_group("marginals")
        for marginal_prob, head in zip(marginal_probs, headers):
            marginal_probs_group.create_dataset(head, data=marginal_prob)


def compute_and_save_ppd(
    model: DistributionLike,
    nf_samples: Array,
    domains: List[Tuple[float, float, int]],
    output_file: str,
    parameters: List[str],
    constants: Dict[str, ArrayLike],
    nf_samples_mapping: Dict[str, int],
    batch_size: int = 1000,
) -> None:
    """Compute the PPD and save it to a file.

    Parameters
    ----------
    model : DistributionLike
        The model.
    nf_samples : Array
        The normalizing flow samples.
    domains : List[Tuple[float, float, int]]
        The domains of the parameters.
    output_file : str
        The file to save the PPD.
    parameters : List[str]
        The parameters of the model.
    constants : Dict[str, ArrayLike]
        The constants of the model.
    nf_samples_mapping : Dict[str, int]
        The mapping of the normalizing flow samples to the model parameters.
    batch_size : int, optional
        The batch size for the computation, defaults to 1000.
    """
    xx = [jnp.linspace(a, b, int(n)) for a, b, n in domains]
    mesh = jnp.meshgrid(*xx, indexing="ij")
    del xx
    xx_mesh = jnp.stack(mesh, axis=-1)
    del mesh
    shape = xx_mesh.shape
    xx_mesh = xx_mesh.reshape(-1, shape[-1])

    prob_values: Array = jax.lax.map(
        partial(
            compute_probs,
            xx_mesh=xx_mesh,
            model=model,
            constants=constants,
            nf_samples_mapping=nf_samples_mapping,
            batch_size=batch_size,
        ),
        nf_samples,
    )
    del nf_samples
    del xx_mesh
    prob_values = prob_values.reshape(-1, *shape[:-1])
    prob_values = np.moveaxis(prob_values, 0, -1)
    ppd_values = np.mean(prob_values, axis=-1)
    marginals = get_all_marginals(prob_values, domains)
    save_probs(ppd_values, marginals, output_file, domains, parameters)



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from ._model import (
    build_eccentricity_distributions as build_eccentricity_distributions,
    build_spin_distributions as build_spin_distributions,
    build_tilt_distributions as build_tilt_distributions,
    NPowerlawMGaussian as NPowerlawMGaussian,
)



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0

#
"""Provides implementation of various constraints using
:class:`~numpyro.distributions.constraints.Constraint`.
"""

from collections.abc import Sequence
from typing import Tuple, Union

from jax import lax, numpy as jnp
from jaxtyping import Array
from numpyro.distributions.constraints import (
    _SingletonConstraint,
    Constraint,
    independent,
    positive,
)
from numpyro.distributions.transforms import Transform


__all__ = [
    "decreasing_vector",
    "increasing_vector",
    "mass_ratio_mass_sandwich",
    "mass_sandwich",
    "positive_decreasing_vector",
    "positive_increasing_vector",
    "strictly_decreasing_vector",
    "strictly_increasing_vector",
    "all_constraint",
    "any_constraint",
]


class _MassSandwichConstraint(Constraint):
    r"""Constrain mass values to lie within a sandwiched interval.

    .. math::
        m_{\text{min}} \leq m_2 \leq m_1 \leq m_{\text{max}}
    """

    event_dim = 1

    def __init__(self, mmin: float, mmax: float):
        """
        Parameters
        ----------
        mmin : float
            Minimum mass.
        mmax : float
            Maximum mass.
        """
        self.mmin = mmin
        self.mmax = mmax

    def __call__(self, x: Array) -> Array:
        m1, m2 = jnp.unstack(x, axis=-1)
        mask = jnp.logical_and(jnp.less(0.0, self.mmin), jnp.less_equal(self.mmin, m2))
        mask = jnp.logical_and(mask, jnp.less_equal(m2, m1))
        mask = jnp.logical_and(mask, jnp.less_equal(m1, self.mmax))
        return jnp.asarray(mask, dtype=bool)

    def feasible_like(self, prototype: Array) -> Array:
        return jnp.full(prototype.shape, (self.mmin + self.mmax) * 0.5)

    def tree_flatten(self):
        return (self.mmin, self.mmax), (("mmin", "mmax"), dict())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, _MassSandwichConstraint):
            return False
        return jnp.array_equal(self.mmin, other.mmin) & jnp.array_equal(
            self.mmax, other.mmax
        )


class _MassRatioMassSandwichConstraint(Constraint):
    r"""Constrain primary mass to lie within a sandwiched interval and the mass ratio to
    lie within a given interval. This is a transformed version of the
    :class:`_MassSandwichConstraint`.

    .. math::
        \begin{align*}
            m_{\text{min}}             & \leq m_1 \leq m_{\max} \\
            \frac{m_{\text{min}}}{m_1} & \leq q   \leq 1
        \end{align*}
    """

    event_dim = 1

    def __init__(self, mmin: float, mmax: float):
        """
        Parameters
        ----------
        mmin : float
            Minimum mass.
        mmax : float
            Maximum mass.
        """
        self.mmin = mmin
        self.mmax = mmax

    def __call__(self, x: Array) -> Array:
        m1, q = jnp.unstack(x, axis=-1)
        m2 = jnp.multiply(m1, q)
        mask = jnp.logical_and(jnp.less(0.0, self.mmin), jnp.less_equal(self.mmin, m2))
        mask = jnp.logical_and(mask, jnp.less_equal(m2, m1))
        mask = jnp.logical_and(mask, jnp.less_equal(m1, self.mmax))
        return jnp.asarray(mask, dtype=bool)

    def feasible_like(self, prototype: Array) -> Array:
        assert prototype.ndim >= 1, "Prototype must have at least one dimension."
        assert prototype.shape[-1] == 2, (
            "Prototype must have last dimension of size 2 for mass and mass ratio."
        )
        shape = prototype.shape[:-1]
        m1 = (self.mmin + self.mmax) * 0.5
        q = jnp.clip(self.mmin / m1, 0.0, 1.0)
        m1 = jnp.broadcast_to(m1, shape)
        q = jnp.broadcast_to(q, shape)
        return jnp.stack((m1, q), axis=-1)

    def tree_flatten(self):
        return (self.mmin, self.mmax), (("mmin", "mmax"), dict())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, _MassRatioMassSandwichConstraint):
            return False
        return jnp.array_equal(self.mmin, other.mmin) & jnp.array_equal(
            self.mmax, other.mmax
        )


class _IncreasingVector(_SingletonConstraint):
    r"""Constrain values to be increasing, i.e. :math:`\forall i<j,x_i\leq x_j`."""

    event_dim = 1

    def __call__(self, x):
        return jnp.all(x[..., 1:] >= x[..., :-1], axis=-1)

    def feasible_like(self, prototype):
        return jnp.ones(prototype.shape, dtype=prototype.dtype)

    def tree_flatten(self):
        return (), ((), dict())

    def __eq__(self, other):
        return isinstance(other, _IncreasingVector)


class _DecreasingVector(_SingletonConstraint):
    r"""Constrain values to be decreasing, i.e. :math:`\forall i<j, x_i \geq x_j`."""

    event_dim = 1

    def __call__(self, x):
        return jnp.all(x[..., 1:] <= x[..., :-1], axis=-1)

    def feasible_like(self, prototype):
        return jnp.ones(prototype.shape, dtype=prototype.dtype)

    def tree_flatten(self):
        return (), ((), dict())

    def __eq__(self, other):
        return isinstance(other, _DecreasingVector)


class _StrictlyIncreasingVector(_SingletonConstraint):
    r"""Constrain values to be strictly increasing, i.e. :math:`\forall i<j, x_i <
    x_j`.
    """

    event_dim = 1

    def __call__(self, x):
        return jnp.all(x[..., 1:] > x[..., :-1], axis=-1)

    def feasible_like(self, prototype):
        return jnp.ones(prototype.shape, dtype=prototype.dtype) * jnp.arange(
            1, prototype.shape[-1] + 1, dtype=prototype.dtype
        )

    def tree_flatten(self):
        return (), ((), dict())

    def __eq__(self, other):
        return isinstance(other, _StrictlyIncreasingVector)


class _StrictlyDecreasingVector(_SingletonConstraint):
    r"""Constrain values to be strictly decreasing, i.e. :math:`\forall i<j,x_i >
    x_j`.
    """

    event_dim = 1

    def __call__(self, x):
        return jnp.all(x[..., 1:] < x[..., :-1], axis=-1)

    def feasible_like(self, prototype):
        return jnp.ones(prototype.shape, dtype=prototype.dtype) * jnp.arange(
            prototype.shape[-1], 0, -1, dtype=prototype.dtype
        )

    def tree_flatten(self):
        return (), ((), dict())

    def __eq__(self, other):
        return isinstance(other, _StrictlyDecreasingVector)


class _PositiveIncreasingVector(_SingletonConstraint):
    r"""Constrain values to be positive and increasing, i.e. :math:`\forall i<j, x_i
    \leq x_j`.
    """

    event_dim = 1

    def __call__(self, x):
        return increasing_vector.check(x) & independent(positive, 1).check(x)

    def feasible_like(self, prototype):
        return jnp.ones(prototype.shape, dtype=prototype.dtype)

    def tree_flatten(self):
        return (), ((), dict())

    def __eq__(self, other):
        return isinstance(other, _PositiveIncreasingVector)


class _PositiveDecreasingVector(_SingletonConstraint):
    r"""Constrain values to be positive and decreasing, i.e. :math:`\forall i<j, x_i
    \geq x_j`.
    """

    event_dim = 1

    def __call__(self, x):
        return decreasing_vector.check(x) & independent(positive, 1).check(x)

    def feasible_like(self, prototype):
        return jnp.ones(prototype.shape, dtype=prototype.dtype)

    def tree_flatten(self):
        return (), ((), dict())

    def __eq__(self, other):
        return isinstance(other, _PositiveDecreasingVector)


class _AllConstraint(Constraint):
    r"""Constrain values to satisfy multiple constraints."""

    event_dim = -1

    def __init__(
        self,
        constraints: Sequence[Constraint],
        event_slices: Sequence[int | Tuple[int, int]],
    ):
        assert len(constraints) == len(event_slices), (
            f"Number of constraints ({len(constraints)}) must match the number of "
            f"event slices ({len(event_slices)})"
        )
        self.constraints = constraints
        self.event_slices = event_slices

    def __call__(self, x):
        mask = None
        for constraint, event_slice in zip(self.constraints, self.event_slices):
            if isinstance(event_slice, int):
                x_slice = lax.dynamic_index_in_dim(
                    x, event_slice, axis=self.event_dim, keepdims=False
                )
            else:
                x_slice = lax.dynamic_slice_in_dim(
                    x,
                    event_slice[0],
                    event_slice[1] - event_slice[0],
                    axis=self.event_dim,
                )
            if mask is None:
                mask = constraint.check(x_slice)
            else:
                mask = jnp.logical_and(mask, constraint.check(x_slice))
        return mask

    def feasible_like(self, prototype: Array) -> Array:
        feasible_values = []
        for constraint, event_slice in zip(self.constraints, self.event_slices):
            if isinstance(event_slice, int):
                prototype_slice = lax.dynamic_index_in_dim(
                    prototype, event_slice, axis=self.event_dim, keepdims=False
                )
            else:
                prototype_slice = lax.dynamic_slice_in_dim(
                    prototype,
                    event_slice[0],
                    event_slice[1] - event_slice[0],
                    axis=self.event_dim,
                )
            feasible_values.append(constraint.feasible_like(prototype_slice))
        max_ndim = prototype.ndim
        feasible_values = [
            jnp.expand_dims(
                feasible_value, axis=tuple(range(feasible_value.ndim, max_ndim))
            )
            if feasible_value.ndim < max_ndim
            else feasible_value
            for feasible_value in feasible_values
        ]
        feasible_value = jnp.concatenate(feasible_values, axis=-1)
        return feasible_value

    def tree_flatten(self):
        return (self.constraints, self.event_slices), (
            ("constraints", "event_slices"),
            dict(),
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, _AllConstraint):
            return False
        return all(
            constraint == other_constraint
            for constraint, other_constraint in zip(self.constraints, other.constraints)
        )


class _AnyConstraint(Constraint):
    r"""Constrain values to satisfy at least one of the constraints."""

    def __init__(self, constraints: Sequence[Constraint]):
        self.constraints = constraints

    def __call__(self, x):
        mask = self.constraints[0].check(x)
        for constraint in self.constraints[1:]:
            mask |= constraint.check(x)
        return mask

    def feasible_like(self, prototype: Array) -> Array:
        return self.constraints[0].feasible_like(prototype)

    def tree_flatten(self):
        return (self.constraints,), (("constraints",), dict())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, _AnyConstraint):
            return False
        return all(
            constraint == other_constraint
            for constraint, other_constraint in zip(self.constraints, other.constraints)
        )


class _TransformedConstraint(Constraint):
    def __init__(
        self,
        base_support: Constraint,
        transforms: Union[Transform, Sequence[Transform]],
    ):
        self.base_support = base_support
        if isinstance(transforms, Transform):
            self.transforms = (transforms,)
        else:
            self.transforms = tuple(transforms)

    def __call__(self, x: Array) -> Array:
        y = x
        mask = jnp.ones(x.shape[:-1], dtype=bool)
        for transform in reversed(self.transforms):
            mask = jnp.logical_and(mask, transform.codomain.check(y))
            y = transform.inv(y)
        mask = jnp.logical_and(mask, self.base_support.check(y))
        return mask

    def feasible_like(self, prototype: Array) -> Array:
        fl = self.base_support.feasible_like(prototype)
        for transform in self.transforms:
            fl = transform(fl)
        return fl

    def tree_flatten(self):
        return (self.base_support, self.transforms), (
            ("base_support", "transforms"),
            dict(),
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, _AnyConstraint):
            return False
        if len(self.transforms) != len(other.transforms):
            return False
        return self.base_support == other.base_support and all(
            transform == other_transform
            for transform, other_transform in zip(self.transforms, other.transforms)
        )


mass_sandwich = _MassSandwichConstraint
mass_ratio_mass_sandwich = _MassRatioMassSandwichConstraint
increasing_vector = _IncreasingVector()
decreasing_vector = _DecreasingVector()
strictly_increasing_vector = _StrictlyIncreasingVector()
strictly_decreasing_vector = _StrictlyDecreasingVector()
positive_increasing_vector = _PositiveIncreasingVector()
positive_decreasing_vector = _PositiveDecreasingVector()
all_constraint = _AllConstraint
any_constraint = _AnyConstraint
transform_constraint = _TransformedConstraint



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from typing import Optional

from jax import numpy as jnp
from jax.typing import ArrayLike
from numpyro.distributions import (
    Beta,
    CategoricalProbs,
    constraints,
    Independent,
    MixtureGeneral,
    MultivariateNormal,
    TruncatedNormal,
    Uniform,
)


def GaussianSpinModel(
    mu_eff: ArrayLike,
    sigma_eff: ArrayLike,
    mu_p: ArrayLike,
    sigma_p: ArrayLike,
    rho: ArrayLike,
    *,
    validate_args: Optional[bool] = None,
) -> MultivariateNormal:
    r"""Bivariate normal distribution for the effective and precessing spins.
    See Eq. (D3) and (D4) in `Population Properties of Compact Objects from
    the Second LIGO-Virgo Gravitational-Wave Transient
    Catalog <https://arxiv.org/abs/2010.14533>`_.

    .. math::
        \left(\chi_{\text{eff}}, \chi_{p}\right) \sim \mathcal{N}\left(
            \begin{bmatrix}
                \mu_{\text{eff}} \\ \mu_{p}
            \end{bmatrix},
            \begin{bmatrix}
                \sigma_{\text{eff}}^2 & \rho \sigma_{\text{eff}} \sigma_{p} \\
                \rho \sigma_{\text{eff}} \sigma_{p} & \sigma_{p}^2
            \end{bmatrix}
        \right)

    where :math:`\chi_{\text{eff}}` is the effective spin and
    :math:`\chi_{\text{eff}}\in[-1,1]` and :math:`\chi_{p}` is the precessing spin and
    :math:`\chi_{p}\in[0,1]`.

    Parameters
    ----------

    mu_eff : ArrayLike
        mean of the effective spin
    sigma_eff : ArrayLike
        standard deviation of the effective spin
    mu_p : ArrayLike
        mean of the precessing spin
    sigma_p : ArrayLike
        standard deviation of the precessing spin
    rho : ArrayLike
        correlation coefficient between the effective and precessing
        spins

    Returns
    -------
    MultivariateNormal
        Multivariate normal distribution for the effective and precessing spins
    """
    return MultivariateNormal(
        loc=jnp.array([mu_eff, mu_p]),
        covariance_matrix=jnp.array(
            [
                [jnp.square(sigma_eff), rho * sigma_eff * sigma_p],
                [rho * sigma_eff * sigma_p, jnp.square(sigma_p)],
            ]
        ),
        validate_args=validate_args,
    )


def IndependentSpinOrientationGaussianIsotropic(
    zeta: ArrayLike,
    scale1: ArrayLike,
    scale2: ArrayLike,
    *,
    validate_args: Optional[bool] = None,
) -> MixtureGeneral:
    r"""A mixture model of spin orientations with isotropic and normally
    distributed components. See Eq. (4) of `Determining the population
    properties of spinning black holes <https://arxiv.org/abs/1704.08370>`_.

    .. math::
        p(z_1,z_2\mid\zeta,\sigma_1,\sigma_2) = \frac{1-\zeta}{4} +
        \zeta\mathbb{I}_{[-1,1]}(z_1)\mathbb{I}_{[-1,1]}(z_2)
        \mathcal{N}(z_1\mid 1,\sigma_1)\mathcal{N}(z_2\mid 1,\sigma_2)

    where :math:`\mathbb{I}(\cdot)` is the indicator function.

    Parameters
    ----------

    zeta : ArrayLike
        The mixing probability of the second component.
    scale1 : ArrayLike
        The standard deviation of the first component.
    scale2 : ArrayLike
        The standard deviation of the second component.

    Returns
    -------
    MixtureGeneral
        Mixture model of spin orientations.
    """
    mixing_probs = jnp.stack([1.0 - zeta, zeta], axis=-1)
    low = -jnp.ones((2,))
    high = jnp.ones((2,))
    batch_dim = 1
    isotropic_component = Independent(
        Uniform(
            low=low,
            high=high,
            validate_args=validate_args,
        ),
        batch_dim,
        validate_args=validate_args,
    )
    gaussian_component = Independent(
        TruncatedNormal(
            loc=high,  # set to high because high=1
            scale=jnp.stack([scale1, scale2], axis=-1),
            low=low,
            high=high,
            validate_args=validate_args,
        ),
        batch_dim,
        validate_args=validate_args,
    )
    return MixtureGeneral(
        mixing_distribution=CategoricalProbs(
            probs=mixing_probs, validate_args=validate_args
        ),
        component_distributions=[isotropic_component, gaussian_component],
        support=constraints.independent(constraints.interval(-1.0, 1.0), batch_dim),
        validate_args=validate_args,
    )


def BetaFromMeanVar(
    mean: ArrayLike,
    variance: ArrayLike,
    *,
    validate_args: Optional[bool] = None,
) -> Beta:
    r"""Beta distribution parameterized by the expected value and variance.

    Parameters
    ----------

    mean : ArrayLike
        Expected value of the beta distribution.
    variance : ArrayLike
        Variance of the beta distribution.
    loc : ArrayLike
        lower bound of the beta distribution, defaults to 0.0
    scale : ArrayLike
        width of the beta distribution, defaults to 1.0

    Returns
    -------
    Beta
        Beta distribution with the specified mean and variance.
    """
    alpha = (jnp.square(mean) * (1 - mean) - mean * variance) / variance
    beta = (mean * jnp.square(1 - mean) - (1 - mean) * variance) / variance
    return Beta(alpha, beta, validate_args=validate_args)



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from collections.abc import Sequence
from typing import Dict, Optional, Union

import h5py
import jax
import numpy as np
from jaxtyping import Array
from loguru import logger

from gwkokab.constants import SECONDS_PER_YEAR
from gwkokab.parameters import Parameters

from ..utils.tools import error_if
from ._abc import VolumeTimeSensitivityInterface


_PARAM_MAPPING = {
    Parameters.PRIMARY_MASS_SOURCE.value: "mass1_source",
    Parameters.PRIMARY_SPIN_X.value: "spin1x",
    Parameters.PRIMARY_SPIN_Y.value: "spin1y",
    Parameters.PRIMARY_SPIN_Z.value: "spin1z",
    Parameters.REDSHIFT.value: "redshift",
    Parameters.SECONDARY_MASS_SOURCE.value: "mass2_source",
    Parameters.SECONDARY_SPIN_X.value: "spin2x",
    Parameters.SECONDARY_SPIN_Y.value: "spin2y",
    Parameters.SECONDARY_SPIN_Z.value: "spin2z",
}


class SemiAnalyticalRealInjectionVolumeTimeSensitivity(VolumeTimeSensitivityInterface):
    injections: Array
    """Array of real injections of shape (n_injections, n_features)."""
    sampling_prob: Array
    """Array of sampling probabilities of shape (n_injections,)."""
    analysis_time_years: float
    """Analysis time in years."""
    total_injections: int
    """Total number of injections.

    This includes both accepted and rejected injections.
    """

    def __init__(
        self,
        parameters: Sequence[str],
        filename: str,
        batch_size: Optional[int] = None,
        far_cut: float = 1.0,
        snr_cut: float = 10.0,
        ifar_pipelines: Optional[Sequence[str]] = None,
    ) -> None:
        """Convenience class for loading a neural vt.

        Parameters
        ----------
        parameters : Sequence[str]
            The names of the parameters that the model expects.
        filename : str
            The filename of the neural vt.
        batch_size: Optional[int]
            The batch size of the neural vt.
        far_cut : float
            The FAR cut to apply to the injections. Default is 1.0.
        snr_cut : float
            The SNR cut to apply to the injections. Default is 10.0.
        ifar_pipelines: Optional[Sequence[str]]
            Pipelines name to select ifar from.
        """
        error_if(not parameters, msg="parameters sequence cannot be empty")
        error_if(
            not isinstance(parameters, Sequence),
            msg=f"parameters must be a Sequence, got {type(parameters)}",
        )
        error_if(
            not all(isinstance(p, str) for p in parameters),
            msg="all parameters must be strings",
        )

        self.batch_size = batch_size

        logger.debug("Loading injection from {}", filename)

        with h5py.File(filename, "r") as f:
            self.analysis_time_years = (
                float(f.attrs["analysis_time_s"]) / SECONDS_PER_YEAR
            )
            logger.debug("Analysis time: {:.2f} years", self.analysis_time_years)

            self.total_injections = int(f.attrs["total_generated"])
            logger.debug("Total injections: {}", self.total_injections)

            injections = f["injections"]

            if ifar_pipelines is None:
                ifar_pipelines = [k for k in injections.keys() if "ifar" in k]
                logger.debug(
                    "No pipelines specified for ifar, using all available: {}",
                    ", ".join(ifar_pipelines),
                )
            else:
                logger.debug(
                    "Selecting ifar from pipelines: {}", ", ".join(ifar_pipelines)
                )

            ifar = np.max([injections[k][:] for k in ifar_pipelines], axis=0)
            snr = injections["optimal_snr_net"][:]

            runs = injections["name"][:].astype(str)
            found = np.where(runs == "o3", ifar > 1 / far_cut, snr > snr_cut)

            n_total = found.shape[0]
            n_found = np.sum(found)
            logger.debug(
                "Found {} out of {} injections with FAR < {} and SNR > {}",
                n_found,
                n_total,
                far_cut,
                snr_cut,
            )

            sampling_prob = (
                injections["sampling_pdf"][found][:]
                / injections["mixture_weight"][found][:]
            )

            χ_1x = injections["spin1x"][found][:]
            χ_1y = injections["spin1y"][found][:]
            χ_1z = injections["spin1z"][found][:]
            χ_2x = injections["spin2x"][found][:]
            χ_2y = injections["spin2y"][found][:]
            χ_2z = injections["spin2z"][found][:]
            a1 = np.sqrt(np.square(χ_1x) + np.square(χ_1y) + np.square(χ_1z))
            a2 = np.sqrt(np.square(χ_2x) + np.square(χ_2y) + np.square(χ_2z))

            injs = []
            for p in parameters:
                if p == Parameters.COS_TILT_1.value:
                    _inj = χ_1z / a1
                elif p == Parameters.COS_TILT_2.value:
                    _inj = χ_2z / a2
                elif p == Parameters.PRIMARY_SPIN_MAGNITUDE.value:
                    _inj = a1
                    # We parameterize spins in spherical coordinates, neglecting azimuthal
                    # parameters. The injections are parameterized in terms of cartesian
                    # spins. The Jacobian is `1 / (2 pi magnitude ** 2)`.
                    logger.debug(
                        "Correcting sampling probability for spherical spin "
                        "parameterization of primary spin."
                    )
                    sampling_prob *= 2.0 * np.pi * np.square(a1)
                elif p == Parameters.SECONDARY_SPIN_MAGNITUDE.value:
                    _inj = a2
                    # We parameterize spins in spherical coordinates, neglecting azimuthal
                    # parameters. The injections are parameterized in terms of cartesian
                    # spins. The Jacobian is `1 / (2 pi magnitude ** 2)`.
                    logger.debug(
                        "Correcting sampling probability for spherical spin "
                        "parameterization of secondary spin."
                    )
                    sampling_prob *= 2.0 * np.pi * np.square(a2)
                else:
                    _inj = injections[_PARAM_MAPPING[p]][found][:]
                injs.append(_inj)

            if (
                Parameters.PRIMARY_SPIN_MAGNITUDE.value not in parameters
                and Parameters.SECONDARY_SPIN_MAGNITUDE.value not in parameters
            ):
                # Eliminating the probability of cartesian spins
                logger.debug(
                    "Eliminating the probability of cartesian spins from sampling probability."
                )
                sampling_prob *= np.square(4.0 * np.pi * a1 * a2)

            self.injections = jax.device_put(np.stack(injs, axis=-1), may_alias=True)

            self.sampling_prob = jax.device_put(sampling_prob, may_alias=True)

            param_ranges: Dict[str, Union[int, float]] = {}

            for i, p in enumerate(parameters):
                try:
                    minimum = f.attrs[p + "_min"]
                except KeyError:
                    minimum = np.min(self.injections[:, i])
                try:
                    maximum = f.attrs[p + "_max"]
                except KeyError:
                    maximum = np.max(self.injections[:, i])

                param_ranges[p + "_min"] = minimum
                param_ranges[p + "_max"] = maximum
            self.parameter_ranges = param_ranges

    def get_logVT(self):
        raise NotImplementedError("Injection based VTs do not have a logVT method.")

    def get_mapped_logVT(self):
        raise NotImplementedError("Injection based VTs do not have a logVT method.")



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import functools as ft
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from typing import Optional

import pandas as pd
from jax import numpy as jnp
from jax.lax import broadcast_shapes
from jaxtyping import Array
from numpyro._typing import DistributionLike
from numpyro.distributions import constraints, Distribution, Independent
from numpyro.distributions.util import promote_shapes, validate_sample

from gwkokab.models import PowerlawPeak, SmoothedTwoComponentPrimaryMassRatio
from gwkokab.models.spin import (
    BetaFromMeanVar,
    IndependentSpinOrientationGaussianIsotropic,
)
from gwkokab.models.utils import (
    JointDistribution,
    ScaledMixture,
)
from gwkokab.parameters import Parameters
from gwkokab.utils.tools import error_if
from kokab.utils import ppd, ppd_parser
from kokab.utils.common import ppd_ranges, read_json


class SimpleRedshiftPowerlaw(Distribution):
    arg_constraints = {
        "z_max": constraints.positive,
        "kappa": constraints.real,
    }
    reparametrized_params = ["z_max", "kappa"]
    pytree_data_fields = ("_support", "kappa", "z_max")

    def __init__(
        self, z_max: Array, kappa: Array, *, validate_args: Optional[bool] = None
    ):
        self.z_max, self.kappa = promote_shapes(z_max, kappa)
        batch_shape = broadcast_shapes(jnp.shape(z_max), jnp.shape(kappa))
        self._support = constraints.interval(0.0, z_max)
        super(SimpleRedshiftPowerlaw, self).__init__(
            batch_shape=batch_shape, validate_args=validate_args
        )

    @constraints.dependent_property(is_discrete=False, event_dim=0)
    def support(self):
        """The support of the distribution, which is the interval [0, z_max]."""
        return self._support

    @validate_sample
    def log_prob(self, value: Array) -> Array:
        r"""Evaluate the psi function at a given redshift.

        .. math::

            \ln\psi(z) = \kappa \log(1 + z)

        Parameters
        ----------
        z : ArrayLike
            Redshift(s) to evaluate.

        Returns
        -------
        ArrayLike
            Values of the psi function.
        """
        return self.kappa * jnp.log1p(value)


def PowerlawPeak_raw(
    use_spin: bool = False,
    use_redshift: bool = False,
    use_tilt: bool = False,
    validate_args: Optional[bool] = None,
    **params: Array,
) -> ScaledMixture:
    smoothing_model = SmoothedTwoComponentPrimaryMassRatio(
        alpha=params["alpha"],
        beta=params["beta"],
        delta=params["delta"],
        lambda_peak=params["lambda_peak"],
        loc=params["loc"],
        mmax=params["mmax"],
        mmin=params["mmin"],
        scale=params["scale"],
        validate_args=validate_args,
    )

    component_distributions = [smoothing_model]

    if use_spin:
        chi_dist = Independent(
            BetaFromMeanVar(
                mean=jnp.stack([params["chi_mean"], params["chi_mean"]], axis=-1),
                variance=jnp.stack(
                    [params["chi_variance"], params["chi_variance"]], axis=-1
                ),
                validate_args=validate_args,
            ),
            reinterpreted_batch_ndims=1,
            validate_args=validate_args,
        )
        component_distributions.append(chi_dist)

    if use_tilt:
        tilt_dist = IndependentSpinOrientationGaussianIsotropic(
            zeta=params["cos_tilt_zeta"],
            scale1=params["cos_tilt_scale"],
            scale2=params["cos_tilt_scale"],
            validate_args=validate_args,
        )

        component_distributions.append(tilt_dist)

    if use_redshift:
        z_max = params["z_max"]
        kappa = params["kappa"]
        powerlaw_z = SimpleRedshiftPowerlaw(
            z_max=z_max, kappa=kappa, validate_args=validate_args
        )

        component_distributions.append(powerlaw_z)

    if len(component_distributions) > 1:
        component_distributions = [
            JointDistribution(*component_distributions, validate_args=validate_args)
        ]

    return ScaledMixture(
        log_scales=jnp.asarray([params["log_rate"]]),
        component_distributions=component_distributions,
        support=component_distributions[0].support,  # type: ignore
        validate_args=validate_args,
    )


def make_parser() -> ArgumentParser:
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser = ppd_parser.get_parser(parser)

    model_group = parser.add_argument_group("Model Options")
    model_group.add_argument(
        "--raw",
        action="store_true",
        help="The raw parameters for this model are primary mass and mass ratio. To"
        "align with the rest of the codebase, we transform primary mass and mass ratio"
        "to primary and secondary mass. This flag will use the raw parameters i.e."
        "primary mass and mass ratio.",
    )
    return parser


def model(raw: bool, **params) -> DistributionLike:
    validate_args = params.pop("validate_args", True)
    _model = PowerlawPeak_raw if raw else PowerlawPeak
    return _model(**params, validate_args=validate_args)


def main() -> None:
    parser = make_parser()
    args = parser.parse_args()

    error_if(
        not str(args.filename).endswith(".hdf5"),
        msg="Output file must be an HDF5 file.",
    )

    constants = read_json(args.constants)
    nf_samples_mapping = read_json(args.nf_samples_mapping)

    use_spin = constants.get("use_spin", False)
    use_tilt = constants.get("use_tilt", False)
    use_redshift = constants.get("use_redshift", False)

    parameters = [Parameters.PRIMARY_MASS_SOURCE.value]
    if args.raw:
        parameters.append(Parameters.MASS_RATIO.value)
    else:
        parameters.append(Parameters.SECONDARY_MASS_SOURCE.value)

    if use_spin:
        parameters.extend(
            [
                Parameters.PRIMARY_SPIN_MAGNITUDE.value,
                Parameters.SECONDARY_SPIN_MAGNITUDE.value,
            ]
        )
    if use_tilt:
        parameters.extend([Parameters.COS_TILT_1.value, Parameters.COS_TILT_2.value])
    if use_redshift:
        parameters.append(Parameters.REDSHIFT.value)

    ranges = ppd_ranges(parameters, args.range)

    nf_samples = pd.read_csv(
        args.sample_filename, delimiter=" ", comment="#", header=None
    ).to_numpy()

    ppd.compute_and_save_ppd(
        ft.partial(model, raw=args.raw),
        nf_samples,
        ranges,
        "rate_scaled_" + args.filename,
        parameters,
        constants,
        nf_samples_mapping,
        args.batch_size,
    )

    nf_samples, constants = ppd.wipe_log_rate(nf_samples, nf_samples_mapping, constants)

    ppd.compute_and_save_ppd(
        ft.partial(model, raw=args.raw),
        nf_samples,
        ranges,
        args.filename,
        parameters,
        constants,
        nf_samples_mapping,
        args.batch_size,
    )



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from argparse import ArgumentParser

from numpyro.distributions.distribution import enable_validation


def get_parser(parser: ArgumentParser) -> ArgumentParser:
    """Populate the command line argument parser with the arguments for the Genie
    script.

    Parameters
    ----------
    parser : ArgumentParser
        Parser to add the arguments to

    Returns
    -------
    ArgumentParser
        the command line argument parser
    """

    # Global enable validation for all distributions
    enable_validation()

    ppd_group = parser.add_argument_group("PPD Options")

    ppd_group.add_argument(
        "--sample-filename",
        help="Path of the file to save the samples.",
        type=str,
        required=True,
    )
    ppd_group.add_argument(
        "--filename",
        help="Path of the file to save the PPD.",
        type=str,
        required=True,
    )
    ppd_group.add_argument(
        "--constants",
        help="Path to the JSON file containing the constant parameters.",
        type=str,
        default="constants.json",
    )
    ppd_group.add_argument(
        "--nf-samples-mapping",
        help="Path to the JSON file containing the mapping of the number of samples.",
        type=str,
        default="nf_samples_mapping.json",
    )
    ppd_group.add_argument(
        "--range",
        help="Range of the PPD for each parameter. The format is 'name min max step'. "
        "Repeat for each parameter.",
        nargs=4,
        action="append",
        type=str,
        required=True,
    )
    ppd_group.add_argument(
        "--batch-size",
        help="Batch size for the computation of log prob per sample.",
        type=int,
        default=1000,
    )

    return parser
